"""batou templating support

Currently we support one templating engine:

Jinja2::
    {% for server in servers %}
    server {{ server.name }}
    {% endfor %}

"""

# There is a weird bug going on with Jinja2 on Python3.6 where this
# test fails if Jinja tries to import ctypes itself ... o_O
import ctypes
import io

import jinja2

from batou import TemplatingError, output


class TemplateEngine(object):
    """Abstract templating wrapper class.

    Use a subclass that connects to a specific template engine.
    """

    @classmethod
    def get(cls, enginename):
        """Return TemplateEngine instance for `enginename`."""
        if enginename.lower() == "jinja2":
            return Jinja2Engine()
        raise NotImplementedError("template engine not known", enginename)

    def template(self, sourcefile, args):
        """Render template from `sourcefile` and return the value."""
        try:
            return self._render_template_file(sourcefile, args).getvalue()
        except jinja2.exceptions.TemplateError as e:
            raise TemplatingError.from_context(e, sourcefile)

    def _render_template_file(self, sourcefile, args):
        """Expand template found in `sourcefile` and return it as StringIO."""
        raise NotImplementedError

    def expand(self, templatestr, args):
        """Expand template in `templatestr` and return it as string."""
        raise NotImplementedError


class Jinja2Engine(TemplateEngine):
    def __init__(self, *args, **kwargs):
        super(Jinja2Engine, self).__init__(*args, **kwargs)
        self.env = jinja2.Environment(
            line_statement_prefix="@@",
            keep_trailing_newline=True,
            undefined=jinja2.StrictUndefined,
        )

    def _render_template_file(self, sourcefile, args):
        with open(sourcefile) as f:
            tmpl = f.read()
        tmpl = self.env.from_string(tmpl)
        output = io.StringIO()
        print(tmpl.render(args), file=output)
        return output

    def expand(self, templatestr, args, identifier="<template>"):
        if len(templatestr) > 100 * 1024:
            output.error(
                "You are trying to render a template that is bigger than "
                "100KiB we've seen that Jinja can crash at large templates "
                "and suggest you find alternatives for this. The affected "
                "template starts with:"
            )
            output.annotate(templatestr[:100])
        try:
            tmpl = self.env.from_string(templatestr)
            tmpl.filename = identifier
            return tmpl.render(**args)
        except Exception as e:
            raise TemplatingError.from_context(e, identifier)



import asyncio
import pickle
import random
import sys
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

from batou import (
    ConfigurationError,
    ReportingException,
    SilentConfigurationError,
)
from batou._output import TerminalBackend, output

from .environment import Environment
from .utils import Timer, locked, notify, self_id


class Connector(threading.Thread):
    def __init__(self, host, sem):
        self.host = host
        self.sem = sem
        self.exc_info = None
        super(Connector, self).__init__(name=host.name)

    def run(self):
        tries = 0
        while True:
            tries += 1
            self.exc_info = None
            self.sem.acquire()
            try:
                self.host.connect()
                break
            except Exception:
                self.exc_info = sys.exc_info()
                if tries >= 3:
                    return
            finally:
                self.sem.release()
            time.sleep(random.randint(1, 2 ** (tries + 1)))

        try:
            self.errors = self.host.start()
        except Exception:
            self.exc_info = sys.exc_info()

    def join(self):
        super(Connector, self).join()
        if self.exc_info:
            exc_type, exc_value, exc_tb = self.exc_info
            raise exc_value.with_traceback(exc_tb)


class ConfigureErrors(ReportingException):
    def __init__(self, errors, all_reporting_hostnames):
        self.errors = errors  # in the format of [(set[reporting_hostnames], set[affected_hostnames], error)]
        self.all_reporting_hostnames = all_reporting_hostnames

    def report(self):
        for reporting_hostnames, affected_hostnames, error in self.errors:
            # if it's a SilentConfigurationError, we don't want to report it
            if isinstance(error, SilentConfigurationError):
                continue
            output.line("")
            # if error is reportable, report otherwise, print traceback
            if hasattr(error, "report"):
                error.report()
            else:
                output.error("Unexpected error:")
                tb = traceback.TracebackException.from_exception(error)
                for line in tb.format():
                    output.line("\t" + line.strip(), red=True)
            if affected_hostnames:
                output.tabular(
                    "Affected hosts", ", ".join(affected_hostnames), red=True
                )
            # if some hosts are not reporting, we want to show them
            if reporting_hostnames != self.all_reporting_hostnames:
                output.tabular(
                    "Reporting hosts", ", ".join(reporting_hostnames), red=True
                )
                output.tabular(
                    "Not reporting",
                    ", ".join(
                        self.all_reporting_hostnames - reporting_hostnames
                    ),
                    red=True,
                )

        output.section(
            "{} ERRORS - CONFIGURATION FAILED".format(len(self.errors)),
            red=True,
        )

    def __str__(self):
        return "{} DeploymentError(s): {}".format(
            len(self.errors), ", ".join(str(e) for e in self.errors)
        )


class Deployment(object):
    _upstream = None

    def __init__(
        self,
        environment,
        platform,
        timeout,
        dirty,
        jobs,
        predict_only=False,
        check_and_predict_local=False,
        provision_rebuild=False,
    ):
        self.environment = Environment(
            environment,
            timeout,
            platform,
            provision_rebuild=provision_rebuild,
            check_and_predict_local=check_and_predict_local,
        )
        self.environment.deployment = self

        self.dirty = dirty
        self.predict_only = predict_only
        self.jobs = jobs

        self.timer = Timer("deployment")

    def load(self):
        output.section("Preparing")

        output.step(
            "main", "Loading environment `{}`...".format(self.environment.name)
        )
        self.environment.load()

        if self.jobs is not None:
            self.jobs = self.jobs
        elif self.environment.jobs is not None:
            self.jobs = int(self.environment.jobs)
        else:
            self.jobs = 1
        output.step("main", "Number of jobs: %s" % self.jobs, debug=True)

        # This is located here to avoid duplicating the verification check
        # when loading the repository on the remote environment object.
        output.step("main", "Verifying repository ...")
        self.environment.repository.verify()

        output.step("main", "Loading secrets ...")
        self.environment.load_secrets()

    def provision(self):
        if not self.environment.provisioners:
            return
        output.section("Provisioning hosts ...")
        for host in self.environment.hosts.values():
            if host.provisioner:
                output.step(
                    host.name,
                    "Provisioning with `{}` provisioner. {}".format(
                        host.provisioner.name,
                        "(Rebuild)" if host.provisioner.rebuild else "",
                    ),
                )
                host.provisioner.provision(host)

    def _connections(self):
        self.environment.prepare_connect()
        hosts = sorted(self.environment.hosts)
        sem = threading.Semaphore(5)
        for i, hostname in enumerate(hosts, 1):
            host = self.environment.hosts[hostname]
            if host.ignore:
                output.step(
                    hostname,
                    "Connection ignored ({}/{})".format(
                        i, len(self.environment.hosts)
                    ),
                    bold=False,
                    red=True,
                )
                continue
            output.step(
                hostname,
                "Connecting via {} ({}/{})".format(
                    self.environment.connect_method,
                    i,
                    len(self.environment.hosts),
                ),
            )
            c = Connector(host, sem)
            c.start()
            yield c

    def connect(self):
        output.section("Connecting hosts and configuring model ...")
        # Consume the connection iterator to start all remaining connections
        # but do not wait for them to be joined.
        with self.timer.step("connect"):
            self.connections = list(self._connections())
            [c.join() for c in self.connections]
        all_errors = []
        all_reporting_hostnames = set()
        for c in self.connections:
            errors = pickle.loads(c.errors)
            reporting_hostname = c.host.name
            all_reporting_hostnames.add(reporting_hostname)
            all_errors.extend([(reporting_hostname, e) for e in errors])
            # collects all errors into (reporting_hostname, error) tuples
        # if there are no connections, then we append a ConfigurationError
        if not self.connections:
            raise ConfigurationError.from_context(
                "No host found in environment."
            )
        # if there are no errors, we're done
        if not all_errors:
            return
        # collect with .should_merge
        errors_by_equivalence_class = []
        for hostname, error in all_errors:
            # check wether it fits into an existing equivalence class
            for equivalence_class in errors_by_equivalence_class:
                # they are non-empty
                if error.should_merge(equivalence_class[0][1]):
                    equivalence_class.append((hostname, error))
                    break
            else:
                # no existing equivalence class fits, create a new one
                errors_by_equivalence_class.append([(hostname, error)])
        # now we have a list of equivalence classes of errors
        # we can merge them into a list of (reporting_hostname, affected_hostnames, error) tuples
        merged_errors = []
        for equivalence_class in errors_by_equivalence_class:
            reporting_hostnames = set(
                hostname for hostname, _ in equivalence_class
            )
            merged_error, affected_hostnames = type(
                equivalence_class[0][1]
            ).merge([e for _, e in equivalence_class])
            merged_errors.append(
                (
                    reporting_hostnames,
                    affected_hostnames,
                    merged_error,
                )
            )

        merged_errors.sort(key=lambda e: getattr(e[2], "sort_key", (-99,)))
        raise ConfigureErrors(merged_errors, all_reporting_hostnames)

    def _launch_components(self, todolist):
        for key, info in list(todolist.items()):
            if info["dependencies"]:
                continue
            del todolist[key]
            asyncio.ensure_future(self._deploy_component(key, info, todolist))

    async def _deploy_component(self, key, info, todolist):
        hostname, component = key
        host = self.environment.hosts[hostname]
        if host.ignore:
            output.step(
                hostname,
                "Skipping component {} ... (Host ignored)".format(component),
                red=True,
            )
        elif info["ignore"]:
            output.step(
                hostname,
                "Skipping component {} ... (Component ignored)".format(
                    component
                ),
                red=True,
            )
        else:
            output.step(
                hostname, "Scheduling component {} ...".format(component)
            )
            await self.loop.run_in_executor(
                None, host.deploy_component, component, self.predict_only
            )

        # Clear dependency from todolist
        for other_component in todolist.values():
            if key in other_component["dependencies"]:
                other_component["dependencies"].remove(key)

        # Trigger start of unblocked dependencies
        self._launch_components(todolist)

    def deploy(self):
        if self.predict_only:
            output.section("Predicting deployment actions")
        else:
            output.section("Deploying")

        with self.timer.step("deploy"):
            # Pick a reference remote (the last we initialised) that will pass us
            # the order we should be deploying components in.
            reference_node = [
                h for h in list(self.environment.hosts.values()) if not h.ignore
            ][0]

            self.loop = asyncio.get_event_loop()
            self.taskpool = ThreadPoolExecutor(self.jobs)
            self.loop.set_default_executor(self.taskpool)
            self._launch_components(reference_node.root_dependencies())

            # asyncio.Task.all_tasks was removed in Python 3.9
            # but the replacement asyncio.all_tasks is only available
            # for Python 3.7 and upwards
            # confer https://docs.python.org/3/whatsnew/3.7.html
            # and https://docs.python.org/3.9/whatsnew/3.9.html
            if sys.version_info < (3, 7):
                all_tasks = asyncio.Task.all_tasks
            else:
                all_tasks = asyncio.all_tasks

            def get_pending():
                return {t for t in all_tasks(self.loop) if not t.done()}

            pending = get_pending()
            while pending:
                self.loop.run_until_complete(asyncio.gather(*pending))
                pending = get_pending()

    def summarize(self):
        output.section("Summary")
        for node in list(self.environment.hosts.values()):
            node.summarize()
        output.annotate(
            f"Deployment took {self.timer.humanize('total', 'connect', 'deploy')}"
        )

    def disconnect(self):
        output.step("main", "Disconnecting from nodes ...", debug=True)
        for node in list(self.environment.hosts.values()):
            node.disconnect()


def main(
    environment,
    platform,
    timeout,
    dirty,
    consistency_only,
    predict_only,
    check_and_predict_local,
    jobs,
    provision_rebuild,
):
    output.backend = TerminalBackend()
    output.line(self_id())
    STEPS = ["load", "provision", "connect", "deploy", "summarize"]
    if consistency_only:
        ACTION = "CONSISTENCY CHECK"
        SUCCESS_FORMAT = {"cyan": True}
        STEPS.remove("deploy")
    elif predict_only:
        ACTION = "DEPLOYMENT PREDICTION"
        SUCCESS_FORMAT = {"purple": True}
    else:
        ACTION = "DEPLOYMENT"
        SUCCESS_FORMAT = {"green": True}
    if check_and_predict_local:
        if (not consistency_only) and (not predict_only):
            output.error(
                "The --local option is only to be used with --consistency-only or --predict-only."
            )
            sys.exit(1)
        ACTION += " (local)"

    with locked(".batou-lock", exit_on_failure=True):
        deployment = Deployment(
            environment,
            platform,
            timeout,
            dirty,
            jobs,
            predict_only,
            check_and_predict_local,
            provision_rebuild,
        )
        environment = deployment.environment
        try:
            for step in STEPS:
                try:
                    getattr(deployment, step)()
                except Exception as e:
                    environment.exceptions.append(e)

                if not environment.exceptions:
                    continue

                # Note: There is a similar sorting / output routine in
                # remote_core __channelexec__. This is a bit of copy/paste
                # due to the way bootstrapping works
                environment.exceptions = list(
                    filter(
                        lambda e: not isinstance(e, SilentConfigurationError),
                        environment.exceptions,
                    )
                )

                environment.exceptions.sort(
                    key=lambda x: getattr(x, "sort_key", (-99,))
                )

                exception = ""
                for exception in environment.exceptions:
                    # if isinstance(exception, ReportingException):
                    # since at least one or two exceptions have to be duck typed:
                    if hasattr(exception, "report"):
                        output.line("")
                        exception.report()
                    else:
                        output.line("")
                        output.error("Unexpected exception")
                        tb = traceback.TracebackException.from_exception(
                            exception
                        )
                        for line in tb.format():
                            output.line("\t" + line.strip(), red=True)

                summary = "{} FAILED (during {})".format(ACTION, step)
                output.section(summary, red=True)

                notify(summary, str(exception))
                sys.exit(1)

        finally:
            deployment.disconnect()
        output.section("{} FINISHED".format(ACTION), **SUCCESS_FORMAT)
        notify("{} SUCCEEDED".format(ACTION), environment.name)



import os.path

from batou import UpdateNeeded, output
from batou.component import Component
from batou.lib.file import Directory, ensure_path_nonexistent


def ensure_empty_directory(path):
    """Ensure path is an empty directory.

    Keep the directory if it exists or create a new one if it
    didn't or if it wasn't a directory.

    This helps in situations where the current workdir may be
    the existing path that needs to be cleaned up.
    """
    if not os.path.isdir(path):
        ensure_path_nonexistent(path)
        os.makedirs(path)
    for file in os.listdir(path):
        ensure_path_nonexistent(os.path.join(path, file))


def exactly_one(*args):
    return sum(map(bool, args)) == 1


class Clone(Component):

    namevar = "url"
    target = "."
    revision = None
    branch = None
    tag = None
    vcs_update = True
    clobber = False

    def configure(self):
        if not exactly_one(self.revision, self.branch, self.tag):
            raise ValueError(
                "Clone(%s) needs exactly one of revision, branch or tag"
                % self.url
            )
        self.target = self.map(self.target)
        self += Directory(self.target)

    def verify(self):
        self._force_clone = False
        if not os.path.exists(self.target):
            self._force_clone = True
            raise UpdateNeeded()
        # if the path does exist but isn't a directory, just let the error
        # bubble for now, the message will at least tell something useful
        with self.chdir(self.target):
            if not os.path.exists(".git"):
                self._force_clone = True
                raise UpdateNeeded()

            if self.remote_url() != self.url:
                self._force_clone = True
                raise UpdateNeeded()

            if not self.vcs_update:
                return

            if self.has_outgoing_changesets():
                output.annotate(
                    "Git clone at {} has outgoing changesets.".format(
                        self.target
                    )
                )

            if self.has_changes():
                if self.clobber:
                    output.annotate(
                        "Git clone at {} is dirty, going to lose changes.".format(
                            self.target
                        ),
                        red=True,
                    )
                    raise UpdateNeeded()
                else:
                    output.annotate(
                        "Git clone at {} is dirty - refusing to clobber. "
                        "Use `clobber=True` if this is intentional .".format(
                            self.target
                        ),
                        red=True,
                    )
                    raise RuntimeError(
                        "Refusing to clobber dirty work directory."
                    )

            if self.revision and self.current_revision() != self.revision:
                raise UpdateNeeded()
            if self.branch and (
                self.current_branch() != self.branch
                or self.has_incoming_changesets()
            ):
                raise UpdateNeeded()
            if self.tag and (
                self.current_tag() != self.tag or self.has_incoming_changesets()
            ):
                raise UpdateNeeded()

    def current_tag(self):
        try:
            with self.chdir(self.target):
                stdout, stderr = self.cmd("LANG=C git describe --tags")
        except RuntimeError:
            return None
        return stdout.strip()

    def current_revision(self):
        try:
            with self.chdir(self.target):
                stdout, stderr = self.cmd("LANG=C git show -s --format=%H")
        except RuntimeError:
            return None
        return stdout.strip()

    def current_branch(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd("git rev-parse --abbrev-ref HEAD")
        return stdout.strip()

    def has_incoming_changesets(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd("git fetch --dry-run")
            return stderr.strip()

    def has_outgoing_changesets(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd("LANG=C git status")
        return "Your branch is ahead of" in stdout

    def has_changes(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd("git status --porcelain")
        return bool(stdout.strip())

    def remote_url(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd("git remote get-url origin")
        return stdout.strip()

    def update(self):
        just_cloned = False
        if self._force_clone:
            ensure_empty_directory(self.target)
            self.cmd(
                self.expand("git clone {{component.url}} {{component.target}}")
            )
            just_cloned = True
        with self.chdir(self.target):
            for filepath in self.untracked_files():
                os.unlink(os.path.join(self.target, filepath))
            if not just_cloned:
                self.cmd("git fetch")
            if self.branch:
                self.cmd(
                    self.expand("git reset --hard origin/{{component.branch}}")
                )
            elif self.tag:
                self.cmd(self.expand("git fetch --tags"))
                self.cmd(self.expand("git reset --hard {{component.tag}}"))
            else:
                self.cmd(self.expand("git reset --hard {{component.revision}}"))

            # XXX We should re-think submodule support; e.g. which revision
            # shall the submodules be updated to?
            self.cmd("git submodule update --init --recursive")

    def untracked_files(self):
        stdout, stderr = self.cmd(
            "git status --porcelain --untracked-files=all"
        )
        items = (line.split(None, 1) for line in stdout.splitlines())
        untracked_items = [
            filepath for status, filepath in items if status == "??"
        ]

        # from the manpage of git-status: If a filename contains whitespace
        # or other nonprintable characters, that field will be quoted
        # in the manner of a C string literal
        # so we need to unquote the filenames
        def unquote_git(s):
            if s.startswith('"') and s.endswith('"'):
                return s[1:-1].encode().decode("unicode-escape")
            return s

        return [unquote_git(item) for item in untracked_items]

    def last_updated(self):
        with self.chdir(self.target):
            if not os.path.exists(".git"):
                return None
            stdout, stderr = self.cmd("git show -s --format=%ct")
            timestamp = stdout.strip()
            return float(timestamp)



"""Platform components specific to gocept.net."""

import os.path

import batou.lib.service
from batou.component import Component, platform
from batou.lib.file import File


@platform("gocept.net", batou.lib.service.Service)
class UserInit(Component):
    """Register a service with user-init."""

    def configure(self):
        self.executable = self.parent.executable
        self.pidfile = self.parent.pidfile
        if not os.path.isabs(self.pidfile):
            self.pidfile = os.path.join(self.root.workdir, self.pidfile)
        target = "/var/spool/init.d/{0}/{1}".format(
            self.host.service_user, os.path.basename(self.parent.executable)
        )
        init_source = os.path.join(
            os.path.dirname(__file__), "resources", "init.sh"
        )
        self += File(target, source=init_source, mode=0o755, leading=True)



import os

import mock
import pytest

from batou.deploy import Deployment
from batou.environment import Environment
from batou.host import RemoteHost
from batou.utils import CmdExecutionError, cmd


@pytest.mark.slow
def test_remote_deployment_initializable(sample_service):
    cmd("hg init")
    with open(".hg/hgrc", "w") as f:
        f.write("[paths]\ndefault=https://example.com")
    env = Environment("test-with-env-config")
    env.load()
    env.configure()
    Deployment(env, platform="", jobs=1, timeout=30, dirty=False)


def test_remote_bundle_breaks_on_missing_head(sample_service):
    cmd("hg init")
    env = mock.Mock()
    env.base_dir = sample_service
    h = RemoteHost("asdf", env)
    from batou.repository import MercurialBundleRepository

    repository = MercurialBundleRepository(env)
    h.rpc = mock.Mock()
    h.rpc.hg_current_heads.return_value = []
    with pytest.raises(ValueError) as e:
        repository.update(h)
    assert e.value.args == (
        "Remote repository did not find any heads. "
        "Can not continue creating a bundle.",
    )


def test_remotehost_start(sample_service):
    env = Environment("test-with-env-config")
    env.load()
    env.configure()
    h = RemoteHost("asdf", env)
    h.connect = mock.Mock()
    h.rpc = mock.Mock()
    h.rpc.ensure_base.return_value = "/tmp"
    h.start()



import os.path

from batou.component import Component, HookComponent
from batou.lib.file import File


def ServiceCheck(description, **kw):
    if kw.pop("nrpe", False):
        return NRPEService(description, **kw)
    return Service(description, **kw)


class Service(HookComponent):

    namevar = "description"
    key = "batou.lib.nagios:Service"

    command = None  # path to executable, if relative then to compdir
    args = ""
    notes_url = ""
    servicegroups = "direct"
    contact_groups = None

    # dependencies are a list of (host_object, check_description)
    depend_on = ()

    def configure(self):
        # ","" are not safe for being referenced in business processes.
        self.description = self.description.replace(",", "_")
        if isinstance(self.contact_groups, (list, tuple)):
            self.contact_groups = ",".join(self.contact_groups)
        super(Service, self).configure()

    @property
    def check_command(self):
        result = self.command
        if self.args:
            result += "!" + self.args
        return result


class NRPEService(Service):

    name = None
    servicegroups = "nrpe"

    def configure(self):
        super(NRPEService, self).configure()
        if not self.name:
            self.name = self.description.lower().replace(" ", "_")

    @property
    def check_command(self):
        return "check_nrpe!%s" % self.name

    @property
    def nrpe_command(self):
        return "{} {}".format(self.command, self.args)


class NagiosServer(Component):
    """Generate a Nagios server config snippet that reflects all active checks.

    The generated config needs to be copied to the Nagios server.

    """

    nagios_cfg = os.path.join(
        os.path.dirname(__file__), "resources", "nagios.cfg"
    )

    # Additional static content that will be appended to the generated file.
    static = ""

    def configure(self):
        self.services = list(self.require(Service.key))
        self.services.sort(key=lambda x: (x.host.name, x.description))

        self += File(
            self.expand("nagios-server-{{host.service_user}}.cfg"),
            source=self.nagios_cfg,
            mode=0o644,
        )


class NRPEHost(Component):
    """Super-component to create an NRPE server config."""

    # XXX gocept-net specific
    nrpe_cfg = os.path.join(os.path.dirname(__file__), "resources", "nrpe.cfg")

    def configure(self):
        self.services = [
            service
            for service in self.require(Service.key, host=self.host)
            if isinstance(service, NRPEService)
        ]
        self.services.sort(key=lambda x: x.name)

        self += File(
            self.expand("/etc/nagios/nrpe/local/" "{{host.service_user}}.cfg"),
            source=self.nrpe_cfg,
            mode=0o644,
        )



from batou.lib.debian import (
    Logrotate,
    LogrotateCronjob,
    RebootCronjob,
    Supervisor,
)
from batou.lib.service import Service

# This is a very rough test to at least ensure we can import the modules
# and trigger configure() runs for the components


def test_rebootcronjob(root):
    service = Service("bobbob")
    cronjob = RebootCronjob()
    root.component += service
    service += cronjob


def test_supervisor(root):
    s = Supervisor()
    root.component += s


def test_logrotatecronjob(root):
    cronjob = LogrotateCronjob()
    root.component += cronjob
    logrotate = Logrotate()
    cronjob += logrotate



import os.path

from batou import UpdateNeeded
from batou.component import Component


class Hello(Component):

    asdf = None

    prov = None
    req = None

    def verify(self):
        self.log("%s verify: asdf=%s", self, self.asdf)
        if not os.path.exists(self.root.name):
            raise UpdateNeeded()

    def update(self):
        with open(self.root.name, "w") as f:
            f.write(self.root.name)


class Hello1(Hello):
    pass


class Hello2(Hello):
    pass


class Hello3(Hello):
    pass


class Hello4(Hello):
    pass


class Hello5(Hello):
    pass


class Hello6(Hello):
    pass


class HelloReq(Component):
    def configure(self):
        self.require("i-need")
        self.log("Pre sub")
        self += Sub()
        self.log("Post sub")


class HelloProv(Component):
    def configure(self):
        self.provide("i-need", self)
        self.log("Provide")


class Sub(Component):
    def configure(self):
        self.log("Sub!")


class Unused(Component):
    pass


class BadUnused(Component):
    def configure(self):
        Unused()


class Component1(Component):
    def configure(self):
        self.hello = self.require_one("key", host=self.host)
        self.require("unrelated")


class Component2(Component):
    def configure(self):
        self.provide("key", "value")



import argparse
import os
import os.path
import sys
import textwrap
from typing import Optional

import importlib_resources

import batou
import batou.deploy
import batou.migrate
import batou.secrets.edit
import batou.secrets.encryption
import batou.secrets.manage
from batou._output import TerminalBackend, output


def main(args: Optional[list] = None) -> None:
    os.chdir(os.environ["APPENV_BASEDIR"])
    version = (
        importlib_resources.files("batou")
        .joinpath("version.txt")
        .read_text()
        .strip()
    )
    parser = argparse.ArgumentParser(
        description=(
            "batou v{}: multi-(host|component|environment|version|platform)"
            " deployment"
        ).format(version),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.set_defaults(func=parser.print_usage)

    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enable debug mode."
    )

    subparsers = parser.add_subparsers()

    # Deploy
    p = subparsers.add_parser("deploy", help="Deploy an environment.")
    p.set_defaults(func=p.print_usage)

    p.add_argument(
        "-p",
        "--platform",
        default=None,
        help="Alternative platform to choose. Empty for no platform.",
    )
    p.add_argument(
        "-t",
        "--timeout",
        default=None,
        help="Override the environment's timeout setting",
    )
    p.add_argument(
        "-D",
        "--dirty",
        action="store_true",
        help="Allow deploying with dirty working copy or outgoing changes.",
    )
    p.add_argument(
        "-c",
        "--consistency-only",
        action="store_true",
        help="Only perform a deployment model and environment "
        "consistency check. Only connects to a single host. "
        "Does not touch anything.",
    )
    p.add_argument(
        "-P",
        "--predict-only",
        action="store_true",
        help="Only predict what updates would happen. "
        "Do not change anything.",
    )
    p.add_argument(
        "-L",
        "--local",
        action="store_true",
        dest="check_and_predict_local",
        help="When running in consistency-only or predict-only mode, "
        "do not connect to the remote host, but check and predict "
        "using the local host's state.",
    )
    p.add_argument(
        "-j",
        "--jobs",
        type=int,
        default=None,
        help="Defines number of jobs running parallel to deploy. "
        "The default results in a serial deployment "
        "of components. Will override the environment settings "
        "for operational flexibility.",
    )
    p.add_argument(
        "--provision-rebuild",
        action="store_true",
        help="Rebuild provisioned resources from scratch. "
        "DANGER: this is potentially destructive.",
    )
    p.add_argument(
        "environment",
        help="Environment to deploy.",
        type=lambda x: x.replace(".cfg", ""),
    )
    p.set_defaults(func=batou.deploy.main)

    # SECRETS
    secrets = subparsers.add_parser(
        "secrets",
        help=textwrap.dedent(
            """
            Manage encrypted secret files. Relies on age (or GPG) being installed and
            configured correctly. """
        ),
    )
    secrets.set_defaults(func=secrets.print_usage)

    sp = secrets.add_subparsers()

    p = sp.add_parser(
        "edit",
        help=textwrap.dedent(
            """
            Encrypted secrets file editor utility. Decrypts file,
            invokes the editor, and encrypts the file again. If called with a
            non-existent file name, a new encrypted file is created.
        """
        ),
    )
    p.set_defaults(func=p.print_usage)

    p.add_argument(
        "--editor",
        "-e",
        metavar="EDITOR",
        default=os.environ.get("EDITOR", "vi"),
        help="Invoke EDITOR to edit (default: $EDITOR or vi)",
    )
    p.add_argument(
        "environment", help="Environment to edit secrets for.", type=str
    )
    p.add_argument(
        "edit_file",
        nargs="?",
        help="Sub-file to edit. (i.e. secrets/{environment}-{subfile}",
    )
    p.set_defaults(func=batou.secrets.edit.main)

    p = sp.add_parser(
        "summary", help="Give a summary of secret files and who has access."
    )
    p.set_defaults(func=batou.secrets.manage.summary)

    p = sp.add_parser(
        "add", help="Add a user's key to one or more secret files."
    )
    p.set_defaults(func=p.print_usage)
    p.add_argument("keyid", help="The user's key ID or email address")
    p.add_argument(
        "--environments",
        default="",
        help="The environments to update. Update all if not specified.",
    )
    p.set_defaults(func=batou.secrets.manage.add_user)

    p = sp.add_parser(
        "remove", help="Remove a user's key from one or more secret files."
    )
    p.set_defaults(func=p.print_usage)
    p.add_argument("keyid", help="The user's key ID or email address")
    p.add_argument(
        "--environments",
        default="",
        help="The environments to update. Update all if not specified.",
    )
    p.set_defaults(func=batou.secrets.manage.remove_user)

    p = sp.add_parser(
        "reencrypt",
        help="Re-encrypt all secret files with the current members.",
    )
    p.set_defaults(func=p.print_usage)
    p.add_argument(
        "--environments",
        default="",
        help="The environments to update. Update all if not specified.",
    )
    p.set_defaults(func=batou.secrets.manage.reencrypt)

    p = sp.add_parser(
        "decrypttostdout",
        help="Decrypt a secret file to stdout, useful for git diff.",
    )
    p.set_defaults(func=p.print_usage)
    p.add_argument(
        "file",
        help="The secret file to decrypt, should be contained in an environment.",
    )
    p.set_defaults(func=batou.secrets.manage.decrypt_to_stdout)

    # migrate
    migrate = subparsers.add_parser(
        "migrate",
        help=textwrap.dedent(
            """
            Migrate the configuration to be compatible with the batou version
            used. Requires to commit the changes afterwards. Might show some
            additional upgrade steps which cannot be performed automatically.
        """
        ),
    )
    migrate.set_defaults(func=migrate.print_usage)
    migrate.add_argument(
        "--bootstrap",
        default=False,
        action="store_true",
        help="Used internally when bootstrapping a new batou project.",
    )
    migrate.set_defaults(func=batou.migrate.main)

    args = parser.parse_args(args)

    # Consume global arguments
    batou.output.enable_debug = args.debug
    batou.secrets.encryption.debug = args.debug
    batou.secrets.manage.debug = args.debug

    # Pass over to function
    if args.func.__name__ == "print_usage":
        args.func()
        sys.exit(1)

    if args.func != batou.migrate.main:
        output.backend = TerminalBackend()
        batou.migrate.assert_up_to_date()

    func_args = dict(args._get_kwargs())
    del func_args["func"]
    del func_args["debug"]
    try:
        return args.func(**func_args)
    except batou.FileLockedError as e:
        # Nicer error reporting for non-deployment commands.
        print(e)



import os.path
import re

from batou import UpdateNeeded, output
from batou.component import Component
from batou.lib.file import Directory
from batou.utils import CmdExecutionError


class Clone(Component):

    namevar = "url"
    target = "."
    revision = None
    branch = None
    vcs_update = True

    _revision_pattern = re.compile(r"changeset: +\d+:([a-f0-9]+)")

    def configure(self):
        if (not self.revision_or_branch) or (self.revision and self.branch):
            raise ValueError(
                "Clone(%s) needs exactly one of revision or branch" % self.url
            )
        self.target = self.map(self.target)
        self += Directory(self.target)

    def verify(self):
        with self.chdir(self.target):
            if not os.path.exists(".hg"):
                raise UpdateNeeded()

            if not self.vcs_update:
                return

            if self.has_outgoing_changesets():
                output.annotate(
                    "Hg clone at {} has outgoing changesets.".format(
                        self.target
                    ),
                    red=True,
                )

            if self.has_changes():
                output.annotate(
                    "Hg clone at {} is dirty, going to lose changes.".format(
                        self.target
                    ),
                    red=True,
                )
                raise UpdateNeeded()

            if self.revision:
                long_rev = len(self.revision) == 40
                if self.current_revision(long_rev) != self.revision:
                    raise UpdateNeeded()
            if self.branch and (
                self.current_branch() != self.branch
                or self.has_incoming_changesets()
            ):
                raise UpdateNeeded()

    @property
    def revision_or_branch(self):
        # Mercurial often takes either a revision or a branch.
        return self.revision or self.branch

    def current_revision(self, long=False):
        debug = "--debug" if long else ""
        stdout, stderr = self.cmd(
            self.expand(
                "LANG=C hg --cwd {{component.target}} {{debug}} parent | "
                "grep changeset:",
                debug=debug,
            )
        )
        match = self._revision_pattern.search(stdout)
        if not match:
            return None
        return match.group(1)

    def current_branch(self):
        stdout, stderr = self.cmd("hg branch")
        return stdout.strip()

    def has_incoming_changesets(self):
        try:
            self.cmd("hg incoming -q -l1")
        except CmdExecutionError as e:
            if e.returncode == 1:
                return False
            raise
        return True

    def has_outgoing_changesets(self):
        try:
            with self.chdir(self.target):
                self.cmd("hg outgoing -q -l1")
        except CmdExecutionError as e:
            if e.returncode == 1:
                return False
            raise
        return True

    def has_changes(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd("hg status")
        return bool(stdout.strip())

    def update(self):
        with self.chdir(self.target):
            if not os.path.exists(".hg"):
                self.cmd(
                    self.expand(
                        "hg clone -u {{component.revision_or_branch}} "
                        "{{component.url}} ."
                    )
                )
                return
            self.cmd(
                self.expand("hg pull --rev {{component.revision_or_branch}}")
            )
            for filepath in self.untracked_files():
                os.unlink(os.path.join(self.target, filepath))
            self.cmd(
                self.expand(
                    "hg update --clean --rev {{component.revision_or_branch}}"
                )
            )

    def untracked_files(self):
        stdout, stderr = self.cmd("hg status -q -u")
        items = (line.split(None, 1) for line in stdout.splitlines())
        return [filepath for status, filepath in items if status == "?"]

    def last_updated(self):
        with self.chdir(self.target):
            if not os.path.exists(".hg"):
                return None
            stdout, stderr = self.cmd(
                'hg log -r %s --template "{date|hgdate}\n"'
                % self.current_revision()
            )
            timestamp, offset = stdout.split()
            return float(timestamp) - float(offset)



import pytest

import batou.component
import batou.lib.service


@pytest.fixture
def deploy_platform():
    return "test"


@pytest.fixture
def service(root, tmpdir, request, deploy_platform):
    old_platforms = getattr(batou.lib.service.Service, "_platforms", {})
    if old_platforms:
        del batou.lib.service.Service._platforms

    root.host.platform = deploy_platform

    @batou.component.platform("test", batou.lib.service.Service)
    class TestPlatformService(batou.component.Component):
        def start(self):
            service._platform_started = True

    @batou.component.platform("nostart", batou.lib.service.Service)
    class NoStartPlatformService(batou.component.Component):
        pass

    batou.lib.service.Service._add_platform("test", TestPlatformService)

    service = batou.lib.service.Service("touch {}/called".format(tmpdir))
    service._platform_started = False
    root.component += service
    root.component.deploy()
    yield service
    batou.lib.service.Service._platforms = old_platforms


def test_start_delegates_to_platform_component(service, tmpdir):
    service.start()
    # The platform specific start() was called
    assert service._platform_started
    # The default implementation was not called
    assert not tmpdir.join("called").check()


@pytest.mark.parametrize("deploy_platform", ["another"])
def test_start_runs_executable_if_no_platform_component(service, tmpdir):
    service.start()
    assert tmpdir.join("called").check()


@pytest.mark.parametrize("deploy_platform", ["nostart"])
def test_start_runs_executable_if_no_platform_specific_start(service, tmpdir):
    service.start()
    assert tmpdir.join("called").check()



import os.path

import pytest

import batou.vfs
from batou import ConfigurationError

from ..cron import CronJob, CronTab, PurgeCronTab


def test_collects_cronjobs_into_crontab(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob("command1", timing="* * * * *")
    root.component += CronJob("command2", timing="* * * * *")
    root.component += CronTab()
    root.component.deploy()
    cron_filename = os.path.join(
        root.environment.workdir_base, "mycomponent/crontab"
    )
    with open(cron_filename) as f:
        crontab = f.read()
    assert "command1" in crontab
    assert "command2" in crontab


def test_purge_crontab_is_empty(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += PurgeCronTab()
    root.component.deploy()
    cron_filename = os.path.join(
        root.environment.workdir_base, "mycomponent/crontab"
    )
    with open(cron_filename) as f:
        crontab = f.read()
    assert crontab == "# Generated by batou. Do not edit.\n\n\n"


def test_empty_crontab_must_raise(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    with pytest.raises(ConfigurationError):
        root.component += CronTab()


def test_non_empty_purge_crontab_must_raise(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob("command1", timing="* * * * *")
    root.component += CronJob("command2", timing="* * * * *")
    with pytest.raises(ConfigurationError):
        root.component += CronTab(purge=True)


def test_additional_envs_can_be_set(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob("command2", timing="* * * * *")
    root.component += CronTab(env=dict(PATH="/foo/bar"))
    root.component.deploy()
    cron_filename = os.path.join(
        root.environment.workdir_base, "mycomponent/crontab"
    )
    with open(cron_filename) as f:
        crontab = f.read()
    assert "PATH=/foo/bar" in crontab



def migrate(output):
    """Manual upgrade steps for 2.3."""
    output(
        "Colliding attributes in environment and secrets",
        """

batou will now fail explicitly if an attribute exists both as an environment
override as well as a secret.

You will usually want to delete it from the environment config as that was
being used previously.

        """,
        "manual",
    )



from batou.utils import get_output


def migrate(output):
    """Manual upgrade steps for 2.3. (updated)"""
    candidates = get_output(
        'grep "Attribute" * -nIr | egrep -v "import"', "<no candidates found>"
    )
    output(
        "`Attribute` defaults now require `ConfigString` for expansion, "
        "mapping, and conversion.",
        f"""

We previously always interpreted strings in attribute defaults as
"config strings" that were passed to the factory/conversion function. However,
there was no way out if the target type was already a string and no
conversion should happen.

You need to review your Attribute usages whether any default strings need to be
encased in `batou.component.ConfigString(x)`.

Candidates:

{candidates}

If you are using `batou_ext`, you will also need to update to a current version
supporting batou 2.3.

        """,
    )



import importlib
import json
import textwrap
from typing import List

import importlib_resources

from batou._output import TerminalBackend, output

CONFIG_FILE_NAME = ".batou.json"
MIGRATION_MODULE = "batou.migrate"


def output_migration_step(
    title: str, text: str, status: str = "manual"
) -> None:
    """Print the information of migration step in a formatted way."""
    if status == "manual":
        label = "🔧"
        args = {"yellow": True}
    elif status == "automatic":
        label = "✅"
        args = {"green": True}
    output.annotate(f"{label} {title}", **args)
    output.line("")
    output.line(textwrap.dedent(text).strip())
    output.line("")


def read_config() -> int:
    """Read the migration configuration file.

    Return the current migration version number.
    Raise a FileNotFoundError in case the migration configuration file is
    missing.
    Raise KeyError if data structure of configuration file is not as expected.
    """
    with open(CONFIG_FILE_NAME) as f:
        return int(json.load(f)["migration"]["version"])


def get_migration_steps() -> List[int]:
    """Return the sorted list of all known migration steps."""
    migration_files = (
        importlib_resources.files(MIGRATION_MODULE)
        .joinpath("migrations")
        .iterdir()
    )
    migration_files = [x.name for x in migration_files]
    return sorted(
        int(x.partition(".")[0])
        for x in migration_files
        if not x.startswith(("_", "tests"))
    )


def migrate(base_version: int) -> int:
    """Run the migration.

    Return the new migration version number.
    """
    steps = [x for x in get_migration_steps() if x > base_version]
    if not steps:
        return base_version
    for step in steps:
        module = importlib.import_module(
            f"{MIGRATION_MODULE}.migrations.{step}"
        )
        output.annotate(f"Version: {step}", bold=True)
        output.line("")
        module.migrate(output_migration_step)
    return step


def write_config(version: int) -> None:
    """Write the version number back the migration configuration file.

    Overwrites already existing configuration file.
    """
    with open(CONFIG_FILE_NAME, "w") as f:
        json.dump({"migration": {"version": version}}, f)
        f.write("\n")


def get_current_version() -> int:
    """Get the version number stored in the configuration file.

    Return `0` if there is no configuration file.
    """
    try:
        version = read_config()
    except FileNotFoundError:
        version = 0
    return version


def get_expected_version() -> int:
    """Get the highest version number deduced from migration files."""
    return get_migration_steps()[-1]


def assert_up_to_date() -> bool:
    """Assert that the current migration version matches the expected one."""
    if get_current_version() == get_expected_version():
        return True
    output.error("Please run `./batou migrate` first.")
    raise SystemExit(-153)


def main(*, bootstrap: bool = False) -> None:
    """Run the ``migrate`` batou subcommand."""
    if bootstrap:
        write_config(get_expected_version())
        return
    output.backend = TerminalBackend()
    base_version = get_current_version()
    output.annotate(f"Current version: {base_version}", bold=True)
    new_version = migrate(base_version)
    if new_version != base_version:
        write_config(new_version)
    output.annotate(f"Reached version: {new_version}", bold=True, green=True)



import os.path

import pytest

import batou.lib.svn
from batou.utils import cmd


@pytest.mark.slow
def test_runs_svn_to_clone_repository(root):
    repos_path = os.path.join(root.environment.workdir_base, "repos")
    cmd("svnadmin create " + repos_path)
    cmd(
        "svn checkout file://{dir} upstream; cd upstream;"
        'touch foo; svn add foo; svn commit -m "bar"'.format(dir=repos_path)
    )
    root.component += batou.lib.svn.Checkout(
        "file://" + repos_path, target="clone", revision="head"
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/foo")
    )
    root.component.deploy()  # trigger verify



from unittest import mock

import pytest

from ..main import main


def test_main__main__1(tmp_path, monkeypatch, capsys):
    """It enforces running `batou migrate` if not on current version."""
    monkeypatch.setenv("APPENV_BASEDIR", str(tmp_path))
    with pytest.raises(SystemExit):
        main(["deploy", "test"])
    assert (
        capsys.readouterr().out
        == "ERROR: Please run `./batou migrate` first.\n"
    )


def test_main__main__2(tmp_path, monkeypatch, capsys):
    """It updates to current version on bootstrap."""
    monkeypatch.setenv("APPENV_BASEDIR", str(tmp_path))
    with mock.patch("batou.migrate.main", spec=True) as migrate_main:
        main(["migrate", "--bootstrap"])
    migrate_main.assert_called_with(bootstrap=True)


def test_main__main__3(tmp_path, monkeypatch, capsys):
    """It has useful usage-message if no arguments are given."""
    monkeypatch.setenv("APPENV_BASEDIR", str(tmp_path))
    calling_args = [
        [],
        ["deploy"],
        ["secrets"],
        ["secrets", "edit"],
        ["secrets", "add"],
        ["secrets", "remove"],
    ]
    for args in calling_args:
        joined = " ".join(args)
        with pytest.raises(SystemExit):
            main(args)
        # should contain "usage:"
        # should contain joined as "usage: <joined> other_stuff"
        std = capsys.readouterr()
        output = std.out if std.out else std.err
        assert output.startswith("usage:")
        assert joined in output



import os.path

import pytest

import batou.lib.supervisor


@pytest.fixture
def supervisor(root, request):
    # Urks. Otherwise OS X ends up with socket paths that are too long.
    supervisor = batou.lib.supervisor.Supervisor(
        pidfile="supervisor.pid", socketpath="/tmp/batou-test-supervisor.sock"
    )
    root.component += supervisor
    root.component.deploy()
    yield supervisor
    supervisor.cmd("{}/bin/supervisorctl shutdown".format(supervisor.workdir))


@pytest.mark.slow
def test_waits_for_start(root, supervisor):
    root.component += batou.lib.supervisor.Program(
        "foo",
        command_absolute=False,
        command="bash",
        args='-c "sleep 1; touch %s/foo; sleep 3600"' % (root.workdir),
        options=dict(startsecs=2),
    )
    root.component.deploy()
    assert os.path.exists("%s/foo" % root.workdir)

    supervisor.cmd("{}/check_supervisor".format(supervisor.workdir))


@pytest.mark.slow
def test_does_not_start_disabled_program(root, supervisor):
    root.component += batou.lib.supervisor.Program(
        "foo",
        command_absolute=False,
        command="bash",
        args='-c "sleep 1; touch %s/foo; sleep 3600"' % (root.workdir),
        options=dict(startsecs=2),
        enable=False,
    )
    root.component.deploy()
    assert not os.path.exists("%s/foo" % root.workdir)


@pytest.mark.slow
def test_program_does_not_start_within_startsecs_raises(root, supervisor):
    root.component += batou.lib.supervisor.Program(
        "foo", command_absolute=False, command="true", options=dict(startsecs=1)
    )
    with pytest.raises(RuntimeError):
        root.component.deploy()


@pytest.mark.slow
def test_starts_stopped_program(root, supervisor):
    root.component += batou.lib.supervisor.Program(
        "foo",
        command_absolute=False,
        command="bash",
        args='-c "sleep 3600"',
        options=dict(startsecs=2),
    )
    root.component.deploy()
    supervisor.cmd("{}/bin/supervisorctl stop foo".format(supervisor.workdir))
    root.component.deploy()
    assert (
        "RUNNING"
        in supervisor.cmd(
            "{}/bin/supervisorctl status foo".format(supervisor.workdir)
        )[0]
    )


@pytest.mark.slow
def test_enable_false_reports_not_running_when_daemon_stopped(root):
    supervisor = batou.lib.supervisor.Supervisor(
        enable=False,
        pidfile="supervisor.pid",
        socketpath="/tmp/batou-test-supervisor.sock",
    )
    root.component += supervisor
    # assert nothing raised
    root.component.deploy()


@pytest.mark.slow
def test_enable_true_reports_not_running_when_daemon_stopped(root, supervisor):
    supervisor.cmd("{}/bin/supervisorctl shutdown".format(supervisor.workdir))
    # assert nothing raised
    root.component.deploy()


def test_setting_nagios_to_True_creates_a_nagios_nrpe_service(root):
    supervisor = batou.lib.supervisor.Supervisor(nagios=True)
    root.component += supervisor
    # assert nothing raised
    root.component.configure()
    assert "NRPEService" in [
        x.__class__.__name__ for x in supervisor.sub_components
    ]



import os.path

import pytest

import batou.lib.mercurial
from batou.utils import cmd


@pytest.fixture(scope="function")
def repos_path(root):
    repos_path = os.path.join(root.environment.workdir_base, "upstream")
    cmd(
        "mkdir {dir}; cd {dir}; hg init;"
        'touch foo; hg add foo; hg commit -m "foo"'.format(dir=repos_path)
    )
    return repos_path


@pytest.mark.slow
def test_runs_hg_to_clone_repository(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", revision="tip"
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/foo")
    )
    root.component.deploy()  # trigger verify


@pytest.mark.slow
def test_setting_branch_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
            dir=repos_path
        )
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/bar")
    )


@pytest.mark.slow
def test_branch_does_switch_branch(root, repos_path):
    cmd(
        'cd {dir}; hg branch bar; hg ci -m "commit branch"'.format(
            dir=repos_path
        )
    )
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="bar"
    )
    root.component.deploy()
    stdout, stderr = cmd(
        "cd {workdir}/clone; hg branch".format(workdir=root.workdir)
    )
    assert "bar" == stdout.strip()


@pytest.mark.slow
def test_set_revision_does_not_pull_when_revision_matches(root, repos_path):
    clone = batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component += clone
    root.component.deploy()
    revision = clone.current_revision()
    clone.revision = revision
    clone.branch = None
    cmd(
        'cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
            dir=repos_path
        )
    )
    root.component.deploy()
    stdout, stderr = cmd(
        "cd {workdir}/clone; LANG=C hg incoming".format(workdir=root.workdir)
    )
    assert "changeset:   1" in stdout


@pytest.mark.slow
def test_set_revision_does_not_change_when_long_revision_matches(
    root, repos_path
):
    clone = batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component += clone
    root.component.deploy()
    stdout, stderr = clone.cmd(
        clone.expand(
            "LANG=C hg --cwd {{component.target}} --debug tip | head -n1"
        )
    )
    _, _, revision = stdout.split(":")
    clone.revision = revision.strip()
    clone.branch = None
    root.component.deploy()
    assert not clone.changed


@pytest.mark.slow
def test_has_changes_counts_changes_to_tracked_files(root, repos_path):
    clone = batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes()
    cmd("touch {}/clone/bar".format(root.workdir))
    cmd("cd {}/clone; hg add bar".format(root.workdir))
    assert clone.has_changes()


@pytest.mark.slow
def test_has_changes_counts_untracked_files_as_changes(root, repos_path):
    clone = batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes()
    cmd("touch {}/clone/bar".format(root.workdir))
    assert clone.has_changes()


@pytest.mark.slow
def test_clean_clone_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
            dir=repos_path
        )
    )
    root.component.deploy()
    assert os.path.isfile(root.component.map("clone/bar"))


@pytest.mark.slow
def test_changes_lost_on_update_with_incoming(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; hg addremove; hg ci -m "commit"'.format(
            dir=repos_path
        )
    )
    cmd("cd {dir}/clone; echo foobar >foo".format(dir=root.workdir))
    root.component.deploy()
    assert os.path.exists(root.component.map("clone/bar"))
    with open(root.component.map("clone/foo")) as f:
        assert not f.read()


@pytest.mark.slow
def test_untracked_files_are_removed_on_update(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component.deploy()
    cmd(
        "cd {dir}/clone; mkdir bar; echo foobar >bar/baz".format(
            dir=root.workdir
        )
    )
    root.component.deploy()
    assert not os.path.exists(root.component.map("clone/bar/baz"))


@pytest.mark.slow
def test_changes_lost_on_update_without_incoming(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default"
    )
    root.component.deploy()
    cmd("cd {dir}/clone; echo foobar >foo".format(dir=root.workdir))
    root.component.deploy()
    with open(root.component.map("clone/foo")) as f:
        assert not f.read()


@pytest.mark.slow
def test_clean_clone_vcs_update_false_leaves_changes_intact(root, repos_path):
    root.component += batou.lib.mercurial.Clone(
        repos_path, target="clone", branch="default", vcs_update=False
    )
    root.component.deploy()
    cmd(
        "cd {dir}; echo foobar >foo; touch bar; hg addremove; "
        'hg ci -m "commit"'.format(dir=repos_path)
    )
    cmd("cd {dir}/clone; echo asdf >foo".format(dir=root.workdir))
    root.component.deploy()
    with open(root.component.map("clone/foo")) as f:
        assert "asdf\n" == f.read()
    assert not os.path.exists(root.component.map("clone/bar"))



import ast
import glob
import json
import os
import os.path
import pathlib
import sys
from configparser import RawConfigParser
from typing import Dict, List, Set

from importlib_metadata import entry_points

import batou.c
import batou.utils
import batou.vfs
from batou import (
    ComponentLoadingError,
    ComponentWithUpdateWithoutVerify,
    ConfigurationError,
    CycleErrorDetected,
    DuplicateHostError,
    DuplicateHostMapping,
    InvalidIPAddressError,
    MissingComponent,
    MissingEnvironment,
    NonConvergingWorkingSet,
    SilentConfigurationError,
    SuperfluousComponentSection,
    SuperfluousSection,
    UnknownComponentConfigurationError,
    UnsatisfiedResources,
    UnusedComponentsInitialized,
    UnusedResources,
)
from batou._output import output
from batou.component import Component, ComponentDefinition, RootComponent
from batou.provision import Provisioner
from batou.repository import Repository
from batou.utils import CycleError, cmd

from .component import load_components_from_file
from .host import Host, LocalHost, RemoteHost
from .resources import Resources
from .secrets import SecretProvider


class UnknownEnvironmentError(ValueError):
    """There is/are no environment(s) for this name(s)."""

    def __init__(self, names: list):
        self.names = names

    def __str__(self):
        return f'Unknown environment(s): {", ".join(self.names)}'


class ConfigSection(dict):
    def as_list(self, option):
        result = self[option]
        if "," in result:
            result = [x.strip() for x in result.split(",")]
        elif "\n" in result:
            result = (x.strip() for x in result.split("\n"))
            result = [x for x in result if x]
        else:
            result = [result]
        return result


class Config(object):
    def __init__(self, path):
        config = RawConfigParser()
        config.optionxform = lambda optionstr: optionstr
        if path:  # Test support
            config.read(path)
        self.config = config

    def __contains__(self, section):
        return self.config.has_section(section)

    def __getitem__(self, section):
        if section not in self:
            raise KeyError(section)
        return ConfigSection(
            (x, self.config.get(section, x))
            for x in self.config.options(section)
        )

    def __iter__(self):
        return iter(self.config.sections())

    def get(self, section, default=None):
        try:
            return self[section]
        except KeyError:
            return default


class Environment(object):
    """An environment assigns components to hosts and provides
    environment-specific configuration for components.
    """

    service_user = None
    require_sudo = None
    host_domain = None
    branch = None
    connect_method = None
    update_method = None
    platform = None
    vfs_sandbox = None
    timeout = None
    target_directory = None
    jobs = None

    require_v4 = True
    require_v6 = False

    repository_url = None
    repository_root = None

    provision_rebuild = False

    host_factory = Host

    def __init__(
        self,
        name,
        timeout=None,
        platform=None,
        basedir=".",
        provision_rebuild=False,
        check_and_predict_local=False,
    ):
        self.name: str = name
        self.hosts: Dict[str, Host] = {}
        self.resources = Resources()
        self.overrides: Dict[str, Dict[str, str]] = {}
        self.secret_data: Set[str] = set()
        self.exceptions: List[Exception] = []
        self.timeout = timeout
        self.platform = platform
        self.provision_rebuild = provision_rebuild
        self.check_and_predict_local = check_and_predict_local

        self.hostname_mapping: Dict[str, str] = {}

        # These are the component classes, decorated with their
        # name.
        self.components: Dict[str, "ComponentDefinition"] = {}
        # These are the components assigned to hosts.
        self.root_components: List[RootComponent] = []

        self.base_dir = os.path.abspath(basedir)
        self.workdir_base = os.path.join(self.base_dir, "work")

        # Additional secrets files as placed in secrets/<env>-<name>
        self.secret_files: Dict[str, str] = {}

        self.provisioners: Dict[str, Provisioner] = {}

    @classmethod
    def all(cls):
        for path in pathlib.Path("environments").glob("*/environment.cfg"):
            yield cls(path.parent.name)

    @classmethod
    def filter(cls, filter):
        if not filter:
            return list(cls.all())
        filter = filter.split(",")
        environments = []
        for e in cls.all():
            if e.name not in filter:
                continue
            filter.remove(e.name)
            environments.append(e)
        if filter:
            raise UnknownEnvironmentError(filter)
        return environments

    def _environment_path(self, path="."):
        return os.path.abspath(
            os.path.join(self.base_dir, "environments", self.name, path)
        )

    def _ensure_environment_dir(self):
        if not os.path.isdir(self._environment_path()):
            os.makedirs(self._environment_path())

    def load(self):
        batou.utils.resolve_override.clear()
        batou.utils.resolve_v6_override.clear()

        config_file = (
            pathlib.Path(self.base_dir)
            / "environments"
            / self.name
            / "environment.cfg"
        )
        if not config_file.exists():
            raise MissingEnvironment.from_context(self)

        mapping_file = self._environment_path("hostmap.json")
        if os.path.exists(mapping_file):
            with open(mapping_file, "r") as f:
                for k, v in json.load(f).items():
                    if k in self.hostname_mapping:
                        raise DuplicateHostMapping.from_context(
                            k, v, self.hostname_mapping[k]
                        )
                    self.hostname_mapping[k] = v

        # Scan all components
        for filename in sorted(
            glob.glob(os.path.join(self.base_dir, "components/*/component.py"))
        ):
            try:
                self.components.update(load_components_from_file(filename))
            except Exception as e:
                exc_type, ex, tb = sys.exc_info()
                self.exceptions.append(
                    ComponentLoadingError.from_context(filename, e, tb)
                )

        config = Config(config_file)

        self.load_environment(config)
        self.load_provisioners(config)
        self.load_hosts(config)
        self.load_resolver(config)

        # load overrides
        for section in config:
            if section.startswith("host:"):
                continue
            if section.startswith("provisioner:"):
                continue
            if not section.startswith("component:"):
                if section not in ["hosts", "environment", "vfs", "resolver"]:
                    self.exceptions.append(
                        SuperfluousSection.from_context(section)
                    )
                continue
            root_name = section.replace("component:", "")
            if root_name not in self.components:
                self.exceptions.append(
                    SuperfluousComponentSection.from_context(root_name)
                )
                continue
            self.overrides.setdefault(root_name, {})
            self.overrides[root_name].update(config[section])

        self.repository = Repository.from_environment(self)

        # The deployment base is the path relative to the
        # repository where batou is located (with ./batou,
        # ./environments, and ./components)
        if self.connect_method == "local":
            self.target_directory = self.repository.root

        self.deployment_base = os.path.relpath(
            self.base_dir, self.repository.root
        )

    def load_secrets(self):
        self.secret_provider = SecretProvider.from_environment(self)
        self.secret_provider.inject_secrets()

    def load_environment(self, config):
        environment = config.get("environment", {})
        for key in [
            "service_user",
            "require_sudo",
            "host_domain",
            "target_directory",
            "connect_method",
            "update_method",
            "branch",
            "platform",
            "timeout",
            "repository_url",
            "repository_root",
            "jobs",
        ]:
            if key not in environment:
                continue
            if getattr(self, key) is not None:
                # Avoid overriding early changes that have already been
                # applied, e.g. by tests.
                continue
            setattr(self, key, environment[key])

        self._set_defaults()

        if self.check_and_predict_local:
            self.connect_method = "local"

        if "vfs" in config:
            sandbox = config["vfs"]["sandbox"]
            sandbox = getattr(batou.vfs, sandbox)(self, config["vfs"])
            self.vfs_sandbox = sandbox

        def parse_addr_option(value):
            if value == "optional":
                return value
            return bool(ast.literal_eval(value))

        for attr in ["require_v4", "require_v6"]:
            if attr not in environment:
                continue
            setattr(self, attr, parse_addr_option(environment[attr]))

        batou.utils.Address.require_v4 = self.require_v4
        batou.utils.Address.require_v6 = self.require_v6

        if self.connect_method == "local":
            self.host_factory = LocalHost
        else:
            self.host_factory = RemoteHost

    def load_resolver(self, config):
        resolver = config.get("resolver", {})
        self._resolve_override = v4 = {}
        self._resolve_v6_override = v6 = {}

        for key, value in list(resolver.items()):
            for ip in value.splitlines():
                ip = ip.strip()
                if not ip:
                    continue
                if "." in ip:
                    v4[key] = ip
                elif ":" in ip:
                    v6[key] = ip
                else:
                    self.exceptions.append(
                        InvalidIPAddressError.from_context(ip)
                    )

        batou.utils.resolve_override.update(v4)
        batou.utils.resolve_v6_override.update(v6)

    def load_provisioners(self, config):
        self.provisioners = {}
        if self.check_and_predict_local:
            return
        for section in config:
            if not section.startswith("provisioner:"):
                continue
            name = section.replace("provisioner:", "")
            method = config[section]["method"]
            factory = entry_points(group="batou.provisioners")[method].load()
            provisioner = factory.from_config_section(name, config[section])
            provisioner.rebuild = self.provision_rebuild
            self.provisioners[name] = provisioner

    def load_hosts(self, config):
        self._load_hosts_single_section(config)
        self._load_hosts_multi_section(config)

        simplified_mapping = {
            k: v for k, v in self.hostname_mapping.items() if k != v
        }
        mapping_file = self._environment_path("hostmap.json")
        if simplified_mapping:
            self._ensure_environment_dir()
            with open(mapping_file, "w") as f:
                json.dump(self.hostname_mapping, f)
        else:
            if os.path.exists(mapping_file):
                os.unlink(mapping_file)

    def _load_hosts_single_section(self, config):
        for literal_hostname in config.get("hosts", {}):
            hostname = literal_hostname.lstrip("!")
            host = self.host_factory(
                hostname,
                self,
                config={
                    "ignore": (
                        "True" if literal_hostname.startswith("!") else "False"
                    )
                },
            )
            self.hosts[host.name] = host
            self._load_host_components(
                host, config["hosts"].as_list(literal_hostname)
            )

    def _load_hosts_multi_section(self, config):
        for section in config:
            if not section.startswith("host:"):
                continue
            hostname = section.replace("host:", "", 1)

            host = self.host_factory(hostname, self, config[section])

            # The name can now have been remapped.
            if host.name in self.hosts:
                self.exceptions.append(
                    DuplicateHostError.from_context(host.name)
                )

            self.hosts[host.name] = host

            self._load_host_components(
                host, config[section].as_list("components")
            )

    def _load_host_components(self, host, component_list):
        components = parse_host_components(component_list)
        for component, settings in list(components.items()):
            try:
                self.add_root(
                    component, host, settings["features"], settings["ignore"]
                )
            except KeyError:
                self.exceptions.append(
                    MissingComponent.from_context(component, host.name)
                )

    def _set_defaults(self):
        if self.update_method is None:
            self.update_method = "rsync"
        if self.connect_method is None:
            self.connect_method = "ssh"

        if self.target_directory is None:
            self.target_directory = "~/deployment"

        if self.platform is None and self.host_domain:
            self.platform = self.host_domain

        if self.timeout is None:
            self.timeout = 3
        else:
            self.timeout = int(self.timeout)

    # API to instrument environment config loading

    def get_host(self, hostname):
        return self.hosts[self.hostname_mapping.get(hostname, hostname)]

    def add_root(self, component_name, host, features=(), ignore=False):
        compdef = self.components[component_name]
        root = RootComponent(
            name=compdef.name,
            environment=self,
            host=host,
            features=features,
            ignore=ignore,
            factory=compdef.factory,
            defdir=compdef.defdir,
            workdir=os.path.join(self.workdir_base, compdef.name),
        )
        self.root_components.append(root)
        return root

    def get_root(self, component_name, host):
        for root in self.root_components:
            if root.host == host and root.name == component_name:
                return root
        raise KeyError(
            "Component {} not configured for host {}".format(
                component_name, host.name
            )
        )

    def prepare_connect(self):
        if self.connect_method == "vagrant":
            output.step("vagrant", "Ensuring machines are up ...")
            cmd("vagrant up")
        elif self.connect_method == "kitchen":
            output.step("kitchen", "Ensuring machines are up ...")
            for fqdn in self.hosts:
                cmd("kitchen create {}".format(fqdn))
            if "BATOU_POST_KITCHEN_CREATE_CMD" in os.environ:
                cmd(
                    "kitchen exec -c '{}'".format(
                        os.environ["BATOU_POST_KITCHEN_CREATE_CMD"]
                    )
                )

    # Deployment API (implements the configure-verify-update cycle)

    def configure(self):
        """Configure all root components.

        Monitor the dependencies between resources and try to reach a stable
        order.

        """
        working_set = set(self.root_components)

        previous_working_sets = []
        exceptions = []
        order = []
        root_dependencies = None

        while working_set:
            exceptions = []
            components_without_verify = []
            previous_working_sets.append(working_set.copy())
            retry = set()
            self.resources.dirty_dependencies.clear()

            for root in working_set:
                try:
                    Component._instances.clear()
                    self.resources.reset_component_resources(root)
                    root.overrides = self.overrides.get(root.name, {})
                    root.prepare()
                except ConfigurationError as e:
                    # A known exception which we can report gracefully later.
                    exceptions.append(e)
                except SilentConfigurationError as e:
                    # A silent exception which we can ignore.
                    pass
                except Exception as e:
                    # An unknown exception which we have to work harder
                    # to report gracefully.
                    ex_type, ex, tb = sys.exc_info()
                    exceptions.append(
                        UnknownComponentConfigurationError.from_context(
                            root, e, tb
                        )
                    )
                else:
                    # warnings: does not fail the deployment
                    # 1. unprepared component warning
                    unprepared_components = []
                    for component in Component._instances:
                        if not component._prepared:
                            unprepared_components.append(component)
                    if unprepared_components:
                        # TODO: activate in future
                        unused_exception = (
                            UnusedComponentsInitialized.from_context(
                                unprepared_components, root
                            )
                        )
                        # exceptions.append(unused_exception)
                        output.warn(str(unused_exception))

                    # 2. a component has .update() but no .verify()
                    def has_original_update_method(component):
                        return type(component).update == Component.update

                    def has_original_verify_method(component):
                        return type(component).verify == Component.verify

                    for component in Component._instances:
                        if (
                            not has_original_update_method(component)
                            and has_original_verify_method(component)
                            and (component not in components_without_verify)
                        ):
                            components_without_verify.append((component, root))
                    # configured this component successfully
                    # we won't have to retry it later
                    continue
                retry.add(root)

            retry.update(self.resources.dirty_dependencies)
            retry.update(self.resources.unsatisfied_components)

            # Try to find a valid order of the components. If we can't then we
            # have detected a dependency cycle and need to stop.

            root_dependencies = self.root_dependencies()
            try:
                order = batou.utils.topological_sort(
                    batou.utils.revert_graph(root_dependencies)
                )
            except CycleError as e:
                exceptions.append(CycleErrorDetected.from_context(e))

            if retry in previous_working_sets:
                # If any resources were required, now is the time to report
                # them.
                if self.resources.unsatisfied:
                    exceptions.append(
                        UnsatisfiedResources.from_context(
                            self.resources.unsatisfied_keys_and_components
                        )
                    )

                cycle_start = previous_working_sets.index(retry)
                # consider all components tried during they cycle as unconfigured
                unconfigured = set().union(*previous_working_sets[cycle_start:])

                # We did not manage to improve on our last working set, so we
                # give up.
                exceptions.append(
                    NonConvergingWorkingSet.from_context(unconfigured)
                )
                break

            working_set = retry

        # warn if a component has .update() but no .verify()
        if components_without_verify:
            component_without_verify_exception = (
                ComponentWithUpdateWithoutVerify.from_context(
                    components_without_verify, root
                )
            )
            output.warn(str(component_without_verify_exception))

        # We managed to converge on a working set. However, some resource were
        # provided but never used. We're rather picky here and report this as
        # an error.
        if self.resources.unused:
            exceptions.append(
                UnusedResources.from_context(self.resources.unused)
            )

        for root in order:
            root.log_finish_configure()

        self.exceptions.extend(exceptions)

        return self.exceptions

    def root_dependencies(self, host=None):
        """Return all roots (host/component) with their direct dependencies.

        This can be used as a "todo" list where all things that have no
        dependencies left can be started to be worked on.

        """
        dependencies = self.resources.get_dependency_graph()
        # Complete the graph with components that do not have any dependencies
        # (yet)
        for root in self.root_components:
            if root not in dependencies:
                dependencies[root] = set()
        if host is not None:
            for root in list(dependencies):
                if root.host.fqdn is not host:
                    del dependencies[root]
        return dependencies

    def map(self, path):
        if self.vfs_sandbox:
            return self.vfs_sandbox.map(path)
        return path

    def components_for(self, host):
        """Return component names for given host name"""
        result = {}
        for component in self.root_components:
            if component.host is host:
                result[component.name] = component
        return result

    def _host_data(self):
        host_data = {}
        for hostname, host in self.hosts.items():
            host_data[hostname] = host.data
        return host_data


def parse_host_components(components):
    """Parse a component list as given in an environment config for a host
    into a dict of dicts:

        {'name': {'features': [], 'ignore': False}}

    If one component is ignored, then the whole set of component features
    is ignored.

    Expected syntax:

    [!]component[:feature], component[:feature]

    """
    result = {}
    for name in components:
        name = name.strip()
        if ":" in name:
            name, feature = name.split(":", 1)
        else:
            feature = None

        ignore = name.startswith("!")
        name = name.lstrip("!")
        result.setdefault(name, {"features": [], "ignore": False})
        result[name]["ignore"] |= ignore
        if feature:
            result[name]["features"].append(feature)
    return result



import pathlib
import sys

from configupdater import ConfigUpdater

from batou import AgeCallError, GPGCallError
from batou.environment import Environment, UnknownEnvironmentError
from batou.secrets.encryption import get_encrypted_file


def summary():
    return_code = 0
    environments = Environment.all()
    for environment in environments:
        print(environment.name)
        try:
            environment.load_secrets()
            environment.secret_provider.summary()
        except GPGCallError as e:
            print(e, file=sys.stderr)
            print(e.output, file=sys.stderr)
            return_code = 1
        except AgeCallError as e:
            print(e, file=sys.stderr)
            print(e.output, file=sys.stderr)
            return_code = 1
    return return_code


def add_user(keyid, environments, **kw):
    """Add a user to given environments.

    If environments is not given, a user is added
    to all environments.

    """
    environments_ = Environment.filter(environments)
    for environment in environments_:
        environment.load_secrets()
        with environment.secret_provider.edit():
            config = environment.secret_provider.config
            members = environment.secret_provider._get_recipients()
            if keyid not in members:
                members.append(keyid)
            config.set("batou", "members", ",\n".join(members).split("\n"))
            environment.secret_provider.write_config(
                str(config).encode("utf-8")
            )


def remove_user(keyid, environments, **kw):
    """Remove a user from a given environment.

    If environments is not given, the user is removed
    from all environments.

    """
    environments_ = Environment.filter(environments)
    for environment in environments_:
        environment.load_secrets()
        with environment.secret_provider.edit():
            config = environment.secret_provider.config
            members = environment.secret_provider._get_recipients()
            if keyid in members:
                members.remove(keyid)
            config.set("batou", "members", ",\n".join(members).split("\n"))
            environment.secret_provider.write_config(
                str(config).encode("utf-8")
            )


def reencrypt(environments, **kw):
    """Re-encrypt all secrets in given environments.

    If environments is not given, all secrets are re-encrypted.

    """
    environments_ = Environment.filter(environments)
    print(f"Re-encrypting environments {[e.name for e in environments_]}")
    for environment in environments_:
        environment.load_secrets()
        with environment.secret_provider.edit():
            config = environment.secret_provider.config
            environment.secret_provider.write_config(
                str(config).encode("utf-8"), force_reencrypt=True
            )


def decrypt_to_stdout(file: str):
    """Decrypt a file and write the content to stdout."""
    file_path = pathlib.Path(file).absolute()
    with get_encrypted_file(file_path) as encrypted_file:
        decrypted = encrypted_file.decrypt()
    if decrypted:
        sys.stdout.buffer.write(decrypted)
    else:
        print("(unable to decrypt content)")
    sys.stdout.flush()
    return 0



import pytest

from batou import (
    CycleErrorDetected,
    NonConvergingWorkingSet,
    UnknownComponentConfigurationError,
    UnsatisfiedResources,
    UnusedResources,
)
from batou.component import Component, ComponentDefinition
from batou.environment import Environment
from batou.host import Host


class Provider(Component):
    def configure(self):
        self.provide("the-answer", 42)


class Consumer(Component):
    def configure(self):
        self.the_answer = self.require("the-answer")


class AggressiveConsumer(Component):
    def configure(self):
        self.the_answer = self.require("the-answer")[0]


class SameHostConsumer(Component):
    def configure(self):
        self.the_answer = self.require("the-answer", self.host)


class Broken(Component):
    def configure(self):
        raise KeyError("foobar")


class Circular1(Component):
    def configure(self):
        self.asdf = self.require("asdf")
        self.provide("bsdf", self)


class Circular2(Component):
    def configure(self):
        self.bsdf = self.require("bsdf")
        self.provide("asdf", self)


class DirtySingularCircularReverse(Component):
    def configure(self):
        self.bsdf = self.require("bsdf", reverse=True, dirty=True)
        self.provide("asdf", self)


@pytest.fixture
def env():
    env = Environment("test")
    for component in list(globals().values()):
        if not isinstance(component, type):
            continue
        if issubclass(component, Component):
            compdef = ComponentDefinition(component)
            env.components[compdef.name] = compdef
    return env


def test_provider_without_consumer_raises_error(env):
    env.add_root("provider", Host("host", env))
    errors = env.configure()
    assert len(errors) == 1
    assert isinstance(errors[0], UnusedResources)


def test_consumer_retrieves_value_from_provider_order1(env):
    provider = env.add_root("provider", Host("test", env))
    consumer = env.add_root("consumer", Host("test", env))
    env.configure()
    assert list(consumer.component.the_answer) == [42]
    assert env.root_dependencies() == {provider: set(), consumer: {provider}}


def test_provider_with_consumer_limited_by_host_raises_error(env):
    env.add_root("provider", Host("test2", env))
    env.add_root("consumer", Host("test2", env))
    env.add_root("samehostconsumer", Host("test", env))
    errors = env.configure()
    assert len(errors) == 2
    assert isinstance(errors[0], UnsatisfiedResources)
    assert isinstance(errors[1], NonConvergingWorkingSet)


def test_consumer_retrieves_value_from_provider_order2(env):
    consumer = env.add_root("consumer", Host("host", env))
    provider = env.add_root("provider", Host("host", env))
    env.configure()
    assert list(consumer.component.the_answer) == [42]
    assert env.root_dependencies() == {provider: set(), consumer: {provider}}


def test_consumer_without_provider_raises_error(env):
    env.add_root("consumer", Host("host", env))
    assert len(env.configure()) > 0
    for exc in env.exceptions:
        if isinstance(exc, UnsatisfiedResources):
            assert set(["the-answer"]) == set(
                [key for key, _, _ in exc.unsatisfied_resources]
            )
            break
    else:
        assert False, "Did not find exception"


def test_aggressive_consumer_raises_unsatisfiedrequirement(env):
    env.add_root("aggressiveconsumer", Host("host", env))
    assert len(env.configure()) > 0
    for exc in env.exceptions:
        if isinstance(exc, UnsatisfiedResources):
            assert set(["the-answer"]) == set(
                [key for key, _, _ in exc.unsatisfied_resources]
            )
            break
    else:
        assert False, "Did not find expected exception."


def test_broken_component_logs_real_exception(env):
    env.add_root("broken", Host("host", env))
    errors = env.configure()
    assert len(errors) == 2
    # could be either one of the two exceptions
    error_types = (UnknownComponentConfigurationError, NonConvergingWorkingSet)
    assert isinstance(errors[0], error_types)
    assert isinstance(errors[1], error_types)


def test_consumer_retrieves_value_from_provider_with_same_host(env):
    host = Host("host", env)
    host2 = Host("host2", env)

    consumer = env.add_root("samehostconsumer", host)
    provider = env.add_root("provider", host)
    consumer2 = env.add_root("samehostconsumer", host2)
    provider2 = env.add_root("provider", host2)
    env.configure()
    assert list(consumer.component.the_answer) == [42]
    assert list(consumer2.component.the_answer) == [42]
    assert env.root_dependencies() == {
        consumer: {provider},
        provider: set(),
        consumer2: {provider2},
        provider2: set(),
    }


def test_components_are_ordered_over_multiple_hosts(env):
    provider1 = env.add_root("provider", Host("host", env))
    provider2 = env.add_root("provider", Host("host2", env))
    consumer1 = env.add_root("consumer", Host("host", env))
    consumer2 = env.add_root("consumer", Host("host2", env))
    env.configure()
    assert env.root_dependencies() == {
        consumer1: {provider1, provider2},
        provider1: set(),
        consumer2: {provider1, provider2},
        provider2: set(),
    }


def test_circular_depending_component(env):
    env.add_root("circular1", Host("test", env))
    env.add_root("circular2", Host("test", env))
    errors = env.configure()
    assert len(errors) == 2
    error_types = (CycleErrorDetected, NonConvergingWorkingSet)
    assert isinstance(errors[0], error_types)
    assert isinstance(errors[1], error_types)


def test_dirty_dependency_for_one_time_retrieval(env, capsys):
    env.add_root("circular1", Host("test", env))
    env.add_root("dirtysingularcircularreverse", Host("test", env))
    env.configure()



import mock
import pytest
from mock import Mock

import batou
import batou.utils
from batou.component import Component, RootComponent
from batou.environment import Config, Environment
from batou.host import Host


def test_environment_should_raise_if_no_config_file(tmpdir):
    e = Environment("foobar")
    with pytest.raises(batou.MissingEnvironment):
        e.load()


def test_load_should_use_defaults(sample_service):
    e = Environment("test-without-env-config")
    e.load()
    assert e.host_domain is None
    assert e.branch is None


def test_load_should_use_config(sample_service):
    assert batou.utils.Address.require_v6 is False
    assert batou.utils.Address.require_v4 is True

    e = Environment("test-with-env-config")
    e.load()
    assert e.service_user == "joe"
    assert e.host_domain == "example.com"
    assert e.branch == "release"

    assert e.hosts["foo1"].service_user == "bob"

    # The keys were loaded into the environment
    assert e.require_v6 is True
    assert e.require_v4 == "optional"

    # The class-based defaults were adapted
    assert batou.utils.Address.require_v6 is True
    assert batou.utils.Address.require_v4 == "optional"


def test_load_ignores_predefined_environment_settings(sample_service):
    e = Environment("test-with-env-config")
    e.service_user = "bob"
    e.host_domain = "sample.com"
    e.branch = "default"
    e.load()
    assert e.service_user == "bob"
    assert e.host_domain == "sample.com"
    assert e.branch == "default"


def test_load_sets_up_overrides(sample_service):
    e = Environment("test-with-overrides")
    e.load()
    assert e.overrides == {"hello1": {"asdf": "1"}}


def test_loads_configures_vfs_sandbox(sample_service):
    e = Environment("test-with-vfs-sandbox")
    e.load()
    assert e.vfs_sandbox.map("/asdf").endswith("work/_/asdf")
    assert e.map("/asdf").endswith("work/_/asdf")


def test_default_sandbox_is_identity():
    e = Environment("foo")
    assert e.map("/asdf") == "/asdf"


def test_get_host_raises_keyerror_if_unknown():
    e = Environment("name")
    with pytest.raises(KeyError):
        e.get_host("asdf")


def test_get_host_without_subdomain_also_works():
    e = Environment("name")
    e.hosts["example.com"] = host = Mock()
    e.host_domain = "example.com"
    assert host == e.get_host("example.com")


def test_get_root_raises_keyerror_on_nonassigned_component():
    e = Environment("foo")
    with pytest.raises(KeyError):
        e.get_root("asdf", Host("localhost", e))


def test_multiple_components(sample_service):
    e = Environment("test-multiple-components")
    e.load()
    components = dict(
        (host, list(sorted(c.name for c in e.root_dependencies(host=host))))
        for host in sorted(e.hosts.keys())
    )
    assert components == dict(
        localhost=["hello1", "hello2"],
        otherhost=["hello3", "hello4"],
        thishost=["hello5", "hello6"],
    )


def test_parse_host_components():
    from batou.environment import parse_host_components

    assert parse_host_components(["asdf"]) == {
        "asdf": {"features": [], "ignore": False}
    }

    assert parse_host_components(["!asdf"]) == {
        "asdf": {"features": [], "ignore": True}
    }

    assert parse_host_components(["!asdf:test", "asdf:bar"]) == {
        "asdf": {"features": ["test", "bar"], "ignore": True}
    }

    assert parse_host_components(["asdf:test", "asdf:bar"]) == {
        "asdf": {"features": ["test", "bar"], "ignore": False}
    }


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_should_merge_single_and_multi_definition(add_root):
    e = Environment("name")
    config = Config(None)
    config.config.read_string(
        """
[hosts]
foo = bar
[host:baz]
components = bar
    """
    )
    e.load_hosts(config)
    assert [
        mock.call("bar", e.hosts["foo"], [], False),
        mock.call("bar", e.hosts["baz"], [], False),
    ] == add_root.call_args_list


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_should_load_single_hosts_section(add_root):
    e = Environment("name")
    config = Config(None)
    config.config.read_string(
        """
[hosts]
foo = bar
    """
    )
    e.load_hosts(config)
    add_root.assert_called_once_with("bar", e.hosts["foo"], [], False)


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_should_load_multi_hosts_section(add_root):
    pass
    e = Environment("name")
    config = Config(None)
    config.config.read_string(
        """
[host:foo]
components = bar
    """
    )
    e.load_hosts(config)
    add_root.assert_called_once_with("bar", e.hosts["foo"], [], False)


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_multi_should_use_ignore_flag(add_root):
    pass
    e = Environment("name")
    config = Config(None)
    config.config.read_string(
        """
[host:foo]
components = bar
ignore = True
    """
    )
    e.load_hosts(config)
    assert e.hosts["foo"].ignore


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_should_break_on_duplicate_definition(add_root):
    e = Environment("name")
    config = Config(None)
    config.config.read_string(
        """
[hosts]
foo = bar
[host:foo]
components = bar
    """
    )
    e.load_hosts(config)
    assert e.exceptions
    assert "foo" == e.exceptions[0].affected_hostname


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_multi_should_use_env_platform(add_root):
    e = Environment("name")
    e.platform = mock.sentinel.platform
    config = Config(None)
    config.config.read_string(
        """
[host:foo]
components = bar
ignore = True
    """
    )
    e.load_hosts(config)
    assert e.hosts["foo"].platform == mock.sentinel.platform


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_multi_should_use_host_platform_if_given(add_root):
    pass
    e = Environment("name")
    e.platform = mock.sentinel.platform
    config = Config(None)
    config.config.read_string(
        """
[host:foo]
components = bar
platform = specific
    """
    )
    e.load_hosts(config)
    assert e.hosts["foo"].platform == "specific"


@mock.patch("batou.environment.Environment.add_root")
def test_load_hosts_single_should_use_env_platform(add_root):
    e = Environment("name")
    e.platform = mock.sentinel.platform
    config = Config(None)
    config.config.read_string(
        """
[hosts]
foo = bar
    """
    )
    e.load_hosts(config)
    assert e.hosts["foo"].platform == mock.sentinel.platform


def test_host_data_is_passed_to_host_object():
    e = Environment("name")
    config = Config(None)
    config.config.read_string(
        """
[host:foo]
components = bar
data-alias = baz
    """
    )
    e.load_hosts(config)
    assert "baz" == e.hosts["foo"].data["alias"]


def test_components_for_host_can_be_retrieved_from_environment(sample_service):
    e = Environment("test-overlapping-components")
    e.load()

    def _get_components(hostname):
        return sorted(e.components_for(host=e.hosts[hostname]).keys())

    assert ["hello1", "hello2"] == _get_components("localhost")
    assert ["hello3"] == _get_components("other")
    assert ["hello2", "hello3"] == _get_components("this")

    localhost = e.hosts["localhost"]
    assert "hello1" in localhost.components
    assert "hello2" in localhost.components
    assert "hello3" not in localhost.components


@mock.patch("batou.remote_core.Output.line")
def test_log_in_component_configure_is_put_out(output, sample_service):
    e = Environment("test-with-provide-require")
    e.load()
    e.configure()
    log = "\n".join(c[0][0].strip() for c in output.call_args_list)
    # Provide is *always* logged first, due to provide/require ordering.
    assert (
        """\
Provide
Pre sub
Sub!
Post sub"""
        == log
    )

    output.reset_mock()
    for root in e.root_dependencies():
        root.component.deploy(True)

    log = "\n".join(c[0][0].strip() for c in output.call_args_list)
    assert (
        """\
localhost > Hello
localhost: <Hello (localhost) "Hello"> verify: asdf=None\
"""
        == log
    )


def test_resolver_overrides(sample_service):
    e = Environment("test-resolver")
    e.load()
    assert {"localhost": "127.0.0.2"} == e._resolve_override
    assert {"localhost": "::2"} == e._resolve_v6_override

    assert "127.0.0.2" == batou.utils.resolve("localhost")
    assert "::2" == batou.utils.resolve_v6("localhost", 0)


def test_resolver_overrides_invalid_address(sample_service):
    e = Environment("test-resolver-invalid")
    e.load()

    errors = e.configure()
    assert len(errors) == 1
    assert "thisisinvalid" == errors[0].address


def test_unused_components_get_reported(sample_service, output):
    e = Environment("test-unused")
    e.load()

    output.backend.output = ""
    e.configure()

    assert "'Unused': BadUnused" in output.backend.output


def test_environment_with_check_local_does_not_load_provisioners(
    sample_service,
):
    e = Environment("test-with-check-local", check_and_predict_local=True)
    e.load()

    assert not e.provisioners



"""Securely edit encrypted secret files."""

import os
import pathlib
import subprocess
import sys
import tempfile
import traceback
from typing import Optional

from batou.environment import Environment

from .encryption import debug

NEW_FILE_TEMPLATE = """\
[batou]
secret_provider = age
members =
"""


class Editor(object):
    def __init__(
        self,
        editor_cmd,
        environment: Environment,
        edit_file: Optional[str] = None,
    ):
        environment.load_secrets()
        self.editor_cmd = editor_cmd
        self.environment = environment
        self.edit_file = edit_file
        self.file = self.environment.secret_provider.edit(edit_file)

    def main(self):
        with self.file:
            self.original_cleartext = self.file.cleartext
            self.cleartext = self.original_cleartext
            if self.file.is_new:
                self.original_cleartext = None
                if self.edit_file:
                    self.cleartext = ""
                else:
                    self.cleartext = NEW_FILE_TEMPLATE

            self.interact()

    def _input(self):
        return input("> ").strip()

    def interact(self):
        cmd = "edit"
        while cmd != "quit":
            try:
                self.process_cmd(cmd)
            except Exception as e:
                print()
                print()
                print(f"An error occurred: {e}")
                print("Traceback:")
                tb = traceback.format_exc()
                tb_lines = tb.splitlines()
                # if tb is too long, only have first and last 10 lines
                if len(tb_lines) > 20 and not debug:
                    print("\n".join(tb_lines[:10]))
                    print("...")
                    print("\n".join(tb_lines[-10:]))
                else:
                    print(tb)
                print()
                print("Your changes are still available. You can try:")
                print("\tedit       -- opens editor with current data again")
                print("\tencrypt    -- tries to encrypt current data again")
                print("\tquit       -- quits and loses your changes")
                cmd = self._input()
            else:
                break

    def process_cmd(self, cmd):
        if cmd == "edit":
            self.edit()
            self.encrypt()
        elif cmd == "encrypt":
            self.encrypt()
        elif cmd == "":
            raise ValueError("empty command")
        else:
            raise ValueError("unknown command `{}`".format(cmd))

    def encrypt(self):
        if self.cleartext == self.original_cleartext:
            print("No changes from original cleartext. Not updating.")
            return
        if self.edit_file:
            # If we're editing a specific file, we can just overwrite it.
            with self.environment.secret_provider.config_file:
                self.environment.secret_provider.write_file(
                    self.file, self.cleartext.encode("utf-8")
                )
        else:
            self.environment.secret_provider.write_config(
                self.cleartext.encode("utf-8")
            )

    def edit(self):
        filename, _encryption_ext = os.path.splitext(self.file.path.name)
        _, suffix = os.path.splitext(filename)
        with tempfile.NamedTemporaryFile(
            prefix="edit", suffix=suffix, mode="w+", encoding="utf-8"
        ) as clearfile:
            clearfile.write(self.cleartext)
            clearfile.flush()

            args = [self.editor_cmd + " " + clearfile.name]

            if debug:
                print("Running editor with command: {}".format(args))

            subprocess.check_call(args, shell=True)

            with open(clearfile.name, "r") as new_clearfile:
                self.cleartext = new_clearfile.read()


def main(editor, environment, edit_file: Optional[str] = None, **kw):
    """Secrets editor console script.

    The main focus here is to avoid having unencrypted files accidentally
    ending up in the deployment repository.

    """

    try:
        editor = Editor(editor, Environment(environment), edit_file)
        editor.main()
    except Exception as e:
        # only print traceback if we're in debug mode
        if debug:
            traceback.print_exc()
        print(e, file=sys.stderr)
        sys.exit(1)



import json
import os
import os.path
import pickle
import pwd
import subprocess
import traceback

# Satisfy flake8 and support testing.
try:
    channel
except NameError:
    channel = None

deployment = None
environment = None
target_directory = ""
deployment_base = ""

# The output class should really live in _output. However, to support
# bootstrapping we define it here and then re-import in the _output module.


class Output(object):
    """Manage the output of various parts of batou to achieve
    consistency wrt to formatting and display.
    """

    enable_debug = False

    def __init__(self, backend):
        self.backend = backend
        self._buffer = []
        self._flushing = False

    # Helpers to allow constructing output with reordering and in a distributed
    # fashion.

    def buffer(self, cmd, *args, **kw):
        self._buffer.append((cmd, args, kw))

    def clear_buffer(self):
        self._buffer.clear()

    def flush_buffer(self):
        if self._flushing:
            return
        self._flushing = True
        for cmd, args, kw in self._buffer:
            getattr(self, cmd)(*args, **kw)
        self.clear_buffer()
        self._flushing = False

    def line(self, message, debug=False, **format):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        self.backend.line(message, **format)

    def annotate(self, message, debug=False, **format):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        lines = message.split("\n")
        message = "\n".join(lines)
        self.line(message, **format)

    def tabular(self, key, value, separator=": ", debug=False, **kw):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        message = f"{key.rjust(15)}{separator}{value}"
        self.annotate(message, **kw)

    def section(self, title, debug=False, **format):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        _format = {"bold": True}
        _format.update(format)
        self.backend.sep("=", title, **_format)

    def sep(self, sep, title, **format):
        self.flush_buffer()
        return self.backend.sep(sep, title, **format)

    def step(self, context, message, debug=False, **format):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        _format = {"bold": True}
        _format.update(format)
        self.line("{}: {}".format(context, message), **_format)

    def error(self, message, exc_info=None, debug=False):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        self.step("ERROR", message, red=True)
        if exc_info:
            if self.enable_debug:
                out = traceback.format_exception(*exc_info)
            else:
                etype, value, _ = exc_info
                out = traceback.format_exception_only(etype, value)
            out = "".join(out)
            out = "      " + out.replace("\n", "\n      ") + "\n"
            self.backend.write(out, red=True)

    def warn(self, message, debug=False):
        if debug and not self.enable_debug:
            return
        self.flush_buffer()
        self.step("WARN", message, yellow=True)


class ChannelBackend(object):
    def __init__(self, channel):
        self.channel = channel

    def _send(self, output_cmd, *args, **kw):
        self.channel.send(("batou-output", output_cmd, args, kw))

    def line(self, message, **format):
        self._send("line", message, **format)

    def sep(self, sep, title, **format):
        self._send("sep", sep, title, **format)

    def write(self, content, **format):
        self._send("write", content, **format)


class Deployment(object):
    """Thin layer to represent a deployment on an agent without
    using the actual deployment class.

    """

    environment = None

    def __init__(
        self,
        env_name,
        host_name,
        overrides,
        secret_files,
        secret_data,
        host_data,
        timeout,
        platform,
        os_env=None,
    ):
        self.env_name = env_name
        self.host_name = host_name
        self.overrides = overrides
        self.host_data = host_data
        self.secret_files = secret_files
        self.secret_data = secret_data
        self.timeout = timeout
        self.platform = platform
        self.os_env = os_env

    def load(self):
        from batou.environment import Environment

        if self.os_env:
            os.environ.update(self.os_env)
        self.environment = Environment(
            self.env_name, self.timeout, self.platform
        )
        self.environment.deployment = self
        self.environment.load()
        self.environment.overrides = self.overrides
        for hostname, data in self.host_data.items():
            self.environment.hosts[hostname].data.update(data)
        self.environment.secret_files = self.secret_files
        self.environment.secret_data = self.secret_data
        return self.environment.configure()

    def deploy(self, root, predict_only):
        host = self.environment.get_host(self.host_name)
        root = self.environment.get_root(root, host)
        root.component.deploy(predict_only)


def lock():
    # XXX implement!
    return "OK"


class CmdError(Exception):
    def __init__(self, cmd, returncode, stdout, stderr):
        self.cmd = cmd
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr

    def report(self):
        output.error(self.cmd)
        output.tabular("Return code", str(self.returncode), red=True)
        output.line("STDOUT", red=True)
        output.annotate(self.stdout.decode("utf-8", errors="replace"))
        output.line("STDERR", red=True)
        output.annotate(self.stderr.decode("utf-8", errors="replace"))


def cmd(c, acceptable_returncodes=[0]):
    process = subprocess.Popen(
        ["LANG=C LC_ALL=C LANGUAGE=C {}".format(c)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        shell=True,
    )
    stdout, stderr = process.communicate()
    # We do not have enough knowledge here to decode so we keep
    # stdout and stderr as byte strings for now.
    if process.returncode not in acceptable_returncodes:
        raise CmdError(c, process.returncode, stdout, stderr)
    return stdout, stderr


def ensure_repository(target, method):
    target = os.path.expanduser(target)
    global target_directory
    target_directory = target

    if not os.path.exists(target):
        os.makedirs(target)

    if method in ["hg-pull", "hg-bundle"]:
        if not os.path.exists(target + "/.hg"):
            cmd("hg init {}".format(target))
    elif method in ["git-pull", "git-bundle"]:
        if not os.path.exists(target + "/.git"):
            cmd("git init {}".format(target))
    elif method in ["rsync", "rsync-ext"]:
        pass
    elif method == "local":
        pass
    else:
        raise RuntimeError("Unknown repository method: {}".format(method))

    return target


def ensure_base(base):
    global deployment_base
    deployment_base = os.path.join(target_directory, base)
    if not os.path.exists(deployment_base):
        os.makedirs(deployment_base)
    return deployment_base


def _hg_current_id():
    output, _ = cmd("hg id -Tjson")
    output = json.loads(output)
    return output[0]["id"]


def hg_current_heads():
    target = target_directory
    os.chdir(target)
    id = _hg_current_id()
    if id == "0000000000000000000000000000000000000000":
        return [id]
    heads, _ = cmd("hg heads -Tjson")
    heads = json.loads(heads)
    return [x["node"] for x in heads]


def hg_pull_code(upstream):
    # TODO Make choice of VCS flexible
    target = target_directory
    os.chdir(target)
    # Phase 1: update working copy
    # XXX manage certificates
    cmd("hg pull {}".format(upstream))


def hg_unbundle_code():
    # TODO Make choice of VCS flexible
    # XXX does this protect us from accidental new heads?
    target = target_directory
    os.chdir(target)
    cmd("hg -y unbundle batou-bundle.hg")


def hg_update_working_copy(branch):
    cmd("hg up -C {}".format(branch))
    return _hg_current_id()


# Unbundle is only called if there actually is something to unbundle (see
# `GitBundleRepository`). Hence setting the origin inside `git_unbundle_code`
# doesn't work very well. OTOH `git_pull_code` is called regardles of the
# repository state. So default to `git-bundle`, and only change for pull.
git_origin = "batou-bundle"


def git_current_head():
    target = target_directory
    os.chdir(target)
    id, err = cmd("git rev-parse HEAD", acceptable_returncodes=[0, 128])
    id = id.strip()
    return id if b"unknown revision" not in err else None


def git_pull_code(upstream, branch):
    global git_origin
    git_origin = "batou-pull"
    target = target_directory
    os.chdir(target)
    out, err = cmd("git remote -v")
    for line in out.splitlines():
        line = line.strip()
        line = line.replace(" ", "\t")
        if not line:
            continue
        name, remote, _ = line.split("\t", 3)
        if name != git_origin:
            continue
        if remote != upstream:
            cmd("git remote remove {origin}".format(origin=git_origin))
        # The batou-pull remote is correctly configured.
        break
    else:
        cmd(
            "git remote add {origin} {upstream}".format(
                origin=git_origin, upstream=upstream
            )
        )
    cmd("git fetch batou-pull")


def git_unbundle_code():
    target = target_directory
    os.chdir(target)
    out, err = cmd("git remote -v")
    if b"batou-bundle" not in out:
        cmd(
            "git remote add {origin} batou-bundle.git".format(origin=git_origin)
        )
    cmd("git fetch {origin}".format(origin=git_origin))


def git_update_working_copy(branch):
    cmd(
        "git reset --hard {origin}/{branch}".format(
            origin=git_origin, branch=branch
        )
    )
    id, _ = cmd("git rev-parse HEAD")
    return id.strip().decode("ascii")


def build_batou():
    os.chdir(deployment_base)
    cmd("./batou --help")


def setup_deployment(*args):
    os.chdir(deployment_base)
    global deployment
    deployment = Deployment(*args)
    errors = deployment.load()
    return pickle.dumps(errors)


def deploy(root, predict_only=False):
    deployment.deploy(root, predict_only)


def root_dependencies():
    deps = {}
    for item in deployment.environment.root_dependencies().items():
        (root, dependencies) = item
        key = (root.host.name, root.name)
        deps[key] = {
            "dependencies": [(r.host.name, r.name) for r in dependencies],
            "ignore": root.ignore,
        }
    return deps


def whoami():
    return pwd.getpwuid(os.getuid()).pw_name


def setup_output(debug):
    from batou._output import output

    output.backend = ChannelBackend(channel)
    output.enable_debug = debug


if __name__ == "__channelexec__":
    output = Output(ChannelBackend(channel))
    while not channel.isclosed():
        task, args, kw = channel.receive()
        # Support slow bootstrapping
        try:
            import batou
        except ImportError:
            pass
        try:
            result = locals()[task](*args, **kw)
            channel.send(("batou-result", result))
        except Exception as e:
            # I voted for duck-typing here as we may be running in the
            # bootstrapping phase and don't have access to all classes yet.
            if hasattr(e, "report"):
                e.report()
                channel.send(("batou-error", None))
            else:
                tb = traceback.format_exc()
                channel.send(("batou-unknown-error", tb))



import os
import pathlib
import sys

import pytest

from batou.environment import Environment
from batou.secrets.edit import Editor
from batou.tests.ellipsis import Ellipsis

from .test_secrets import encrypted_file


def test_edit_gpg(tmpdir):
    editor = Editor(
        "true",
        environment=Environment(
            "tutorial", basedir="examples/tutorial-secrets"
        ),
        edit_file="asdf",
    )
    editor.cleartext = "asdf"
    editor.edit()
    assert editor.cleartext == "asdf"


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="age is available in tests with python 3.7 only",
)
def test_edit_age(tmpdir):
    editor = Editor(
        "true",
        environment=Environment("age", basedir="examples/tutorial-secrets"),
        edit_file="asdf",
    )
    editor.cleartext = "asdf"
    editor.edit()
    assert editor.cleartext == "asdf"


def test_edit_command_loop(tmpdir, capsys):
    editor = Editor(
        "true",
        environment=Environment(
            "tutorial", basedir="examples/tutorial-secrets"
        ),
    )
    editor.cleartext = "asdf"

    with pytest.raises(ValueError):
        editor.process_cmd("asdf")

    def broken_cmd():
        raise RuntimeError("gpg is broken")

    editor.edit = broken_cmd
    editor.encrypt = broken_cmd

    cmds = ["edit", "asdf", "encrypt", "quit"]

    def _input():
        return cmds.pop(0)

    editor._input = _input
    editor.interact()

    out, err = capsys.readouterr()
    assert err == ""
    assert out == Ellipsis(
        """\


An error occurred: gpg is broken
Traceback:
Traceback (most recent call last):
  File ".../src/batou/secrets/edit.py", line ..., in interact
    self.process_cmd(cmd)
  File ".../src/batou/secrets/edit.py", line ..., in process_cmd
    self.edit()
  File ".../src/batou/secrets/tests/test_editor.py", line ..., in broken_cmd
    raise RuntimeError("gpg is broken")
RuntimeError: gpg is broken


Your changes are still available. You can try:
\tedit       -- opens editor with current data again
\tencrypt    -- tries to encrypt current data again
\tquit       -- quits and loses your changes


An error occurred: gpg is broken
Traceback:
Traceback (most recent call last):
  File ".../src/batou/secrets/edit.py", line ..., in interact
    self.process_cmd(cmd)
  File ".../src/batou/secrets/edit.py", line ..., in process_cmd
    self.edit()
  File ".../src/batou/secrets/tests/test_editor.py", line ..., in broken_cmd
    raise RuntimeError("gpg is broken")
RuntimeError: gpg is broken


Your changes are still available. You can try:
\tedit       -- opens editor with current data again
\tencrypt    -- tries to encrypt current data again
\tquit       -- quits and loses your changes


An error occurred: unknown command `asdf`
Traceback:
Traceback (most recent call last):
  File ".../src/batou/secrets/edit.py", line ..., in interact
    self.process_cmd(cmd)
  File ".../src/batou/secrets/edit.py", line ..., in process_cmd
    raise ValueError("unknown command `{}`".format(cmd))
ValueError: unknown command `asdf`


Your changes are still available. You can try:
\tedit       -- opens editor with current data again
\tencrypt    -- tries to encrypt current data again
\tquit       -- quits and loses your changes


An error occurred: gpg is broken
Traceback:
Traceback (most recent call last):
  File ".../src/batou/secrets/edit.py", line ..., in interact
    self.process_cmd(cmd)
  File ".../src/batou/secrets/edit.py", line ..., in process_cmd
    self.encrypt()
  File ".../src/batou/secrets/tests/test_editor.py", line ..., in broken_cmd
    raise RuntimeError("gpg is broken")
RuntimeError: gpg is broken


Your changes are still available. You can try:
\tedit       -- opens editor with current data again
\tencrypt    -- tries to encrypt current data again
\tquit       -- quits and loses your changes
"""
    )


def test_edit_file_has_secret_prefix_gpg(tmpdir, encrypted_file):
    filename = "asdf123"
    editor = Editor(
        "true",
        environment=Environment(
            "tutorial", basedir="examples/tutorial-secrets"
        ),
        edit_file=filename,
    )
    assert editor.file.path.name == f"secret-{filename}.gpg"


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="age is available in tests with python 3.7 only",
)
def test_edit_file_has_secret_prefix_age(tmpdir, encrypted_file):
    filename = "asdf123"
    editor = Editor(
        "true",
        environment=Environment("age", basedir="examples/tutorial-secrets"),
        edit_file=filename,
    )
    assert editor.file.path.name == f"secret-{filename}.age"


def test_blank_edit(tmpdir, encrypted_file):
    editor = Editor(
        "true",
        environment=Environment(
            "tutorial", basedir="examples/tutorial-secrets"
        ),
        edit_file="asdf",
    )
    with editor.file:
        editor.original_cleartext = None
        editor.cleartext = ""
        editor.edit()
        assert editor.cleartext == ""
        editor.encrypt()
    with open(editor.file.path, "rb") as f:
        assert f.read() != b""
    os.unlink(editor.file.path)



import os.path

import pytest

import batou.utils


@pytest.fixture(autouse=True)
def ensure_gpg_homedir(monkeypatch):
    home = os.path.join(
        os.path.dirname(__file__), "secrets", "tests", "fixture", "gnupg"
    )
    monkeypatch.setitem(os.environ, "GNUPGHOME", home)


@pytest.fixture(autouse=True)
def ensure_age_identity(monkeypatch):
    key = os.path.join(
        os.path.dirname(__file__),
        "secrets",
        "tests",
        "fixture",
        "age",
        "id_ed25519",
    )
    monkeypatch.setitem(os.environ, "BATOU_AGE_IDENTITIES", key)


@pytest.fixture(autouse=True)
def reset_address_defaults():
    v4, v6 = batou.utils.Address.require_v4, batou.utils.Address.require_v6
    yield
    batou.utils.Address.require_v4, batou.utils.Address.require_v6 = v4, v6



import sys

from batou import (
    ComponentLoadingError,
    ConfigurationError,
    ConversionError,
    CycleErrorDetected,
    DuplicateComponent,
    DuplicateHostError,
    InvalidIPAddressError,
    MissingComponent,
    MissingEnvironment,
    MissingOverrideAttributes,
    NonConvergingWorkingSet,
    SuperfluousComponentSection,
    SuperfluousSecretsSection,
    SuperfluousSection,
    UnknownComponentConfigurationError,
    UnsatisfiedResources,
    UnusedResources,
)
from batou.component import ComponentDefinition


def test_configurationerrors_can_be_sorted(root):
    errors = []
    errors.append(ConfigurationError.from_context("asdffdas", root.component))
    errors.append(
        ConversionError.from_context(
            root.component, "testkey", "testvalue", str, "foobar"
        )
    )
    errors.append(
        MissingOverrideAttributes.from_context(
            root.component, ["asdf", "bsdfg"]
        )
    )

    errors.append(
        DuplicateComponent.from_context(
            ComponentDefinition(root.component.__class__, "asdf.py"),
            ComponentDefinition(root.component.__class__, "bsdf.py"),
        )
    )

    try:
        raise ValueError("asdf")
    except Exception:
        _, exc_value, exc_traceback = sys.exc_info()

    errors.append(
        UnknownComponentConfigurationError.from_context(
            root, exc_value, exc_traceback
        )
    )

    errors.append(UnusedResources.from_context({"asdf": {root: 1}}))

    errors.append(UnsatisfiedResources.from_context({("asdf", None): [root]}))

    errors.append(MissingEnvironment.from_context(root.environment))

    errors.append(
        ComponentLoadingError.from_context("asdf.py", ValueError("asdf"), None)
    )

    errors.append(MissingComponent.from_context("component", "hostname"))

    errors.append(SuperfluousSection.from_context("asdf"))

    errors.append(SuperfluousComponentSection.from_context("asdf"))

    errors.append(SuperfluousSecretsSection.from_context("asdf"))

    errors.append(CycleErrorDetected.from_context("foo"))

    errors.append(NonConvergingWorkingSet.from_context([root]))

    errors.append(DuplicateHostError.from_context("asdf"))

    errors.append(InvalidIPAddressError.from_context(("127.0.0.256/24")))

    errors.sort(key=lambda x: x.sort_key)



import os.path

import importlib_resources

from batou.component import Component, HookComponent, platform
from batou.lib.file import File


class RotatedLogfile(HookComponent):

    namevar = "path"
    key = "batou.lib.logrotate:RotatedLogfile"

    args = ""
    prerotate = None
    postrotate = None

    def configure(self):
        super(RotatedLogfile, self).configure()
        self.path = os.path.join(self.workdir, self.path)
        self.path = self.map(self.path)
        self.args = list(map(str.strip, self.args.split(",")))


class Logrotate(Component):

    common_config = b""
    logrotate_template = (
        importlib_resources.files(__name__)
        .joinpath("resources/logrotate.in")
        .read_bytes()
    )

    def configure(self):
        self.logfiles = self.require(RotatedLogfile.key, host=self.host)
        self.logfiles.sort(key=lambda logfile: logfile.path)

        config = self.common_config + self.logrotate_template
        self.logrotate_conf = File("logrotate.conf", content=config)
        self += self.logrotate_conf


@platform("gocept.net", Logrotate)
class GoceptNetRotatedLogrotate(Component):
    def configure(self):
        user = self.host.service_user
        user_logrotate_conf = os.path.join("/var/spool/logrotate/", user)
        self += File(
            user_logrotate_conf,
            ensure="symlink",
            link_to=self.parent.logrotate_conf.path,
        )



import os
import os.path
import shutil

from batou.environment import Environment
from batou.tests.ellipsis import Ellipsis
from batou.utils import cmd


def test_service_early_resource():
    env = Environment(
        "dev",
        basedir=os.path.dirname(__file__) + "/fixture/service_early_resource",
    )
    env.load()
    env.configure()
    assert env.resources.get("zeo") == ["127.0.0.1:9000"]


def test_example_errors_early():
    os.chdir("examples/errors")
    out, _ = cmd("./batou deploy errors", acceptable_returncodes=[1])
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `errors`...
main: Verifying repository ...
main: Loading secrets ...

ERROR: Failed loading component file
           File: .../examples/errors/components/component5/component.py
      Exception: invalid syntax (component.py, line 1)
Traceback (simplified, most recent call last):
<no-non-remote-internal-traceback-lines-found>

ERROR: Failed loading component file
           File: .../examples/errors/components/component6/component.py
      Exception: No module named 'asdf'
Traceback (simplified, most recent call last):
  File ".../errors/components/component6/component.py", line 1, in <module>
    import asdf  # noqa: F401 import unused
...
ERROR: Missing component
      Component: missingcomponent

ERROR: Superfluous section in environment configuration
        Section: superfluoussection

ERROR: Override section for unknown component found
      Component: nonexisting-component-section

ERROR: Attribute override found both in environment and secrets
      Component: component1
      Attribute: my_attribute

ERROR: Secrets section for unknown component found
      Component: another-nonexisting-component-section
======================= DEPLOYMENT FAILED (during load) ========================
"""
    )  # noqa: E501 line too long


def test_example_errors_gpg_cannot_decrypt(monkeypatch):
    monkeypatch.setitem(os.environ, "GNUPGHOME", "")
    os.chdir("examples/errors")
    out, _ = cmd("./batou deploy errors", acceptable_returncodes=[1])
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `errors`...
main: Verifying repository ...
main: Loading secrets ...

ERROR: Error while calling GPG
        command: gpg --decrypt ...environments/errors/secrets.cfg.gpg
      exit code: 2
        message:
gpg: ...
...
ERROR: Failed loading component file
           File: .../examples/errors/components/component5/component.py
      Exception: invalid syntax (component.py, line 1)
Traceback (simplified, most recent call last):
<no-non-remote-internal-traceback-lines-found>

ERROR: Failed loading component file
           File: .../examples/errors/components/component6/component.py
      Exception: No module named 'asdf'
Traceback (simplified, most recent call last):
  File ".../examples/errors/components/component6/component.py", line 1, in <module>
    import asdf  # noqa: F401 import unused
...
ERROR: Missing component
      Component: missingcomponent

ERROR: Superfluous section in environment configuration
        Section: superfluoussection

ERROR: Override section for unknown component found
      Component: nonexisting-component-section
======================= DEPLOYMENT FAILED (during load) ========================
"""
    )  # noqa: E501 line too long


def test_example_errors_late():
    os.chdir("examples/errors2")
    out, _ = cmd("./batou deploy errors", acceptable_returncodes=[1])
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `errors`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/1)


ERROR: Trying to access address family IPv6 which is not configured for localhost:22.
           Hint: Use `require_v6=True` when instantiating the Address object.

ERROR: crontab: No cron jobs found.
 Affected hosts: localhost

ERROR: malformed node or strin...: <...Name object at 0x...>
      Attribute: Component1.do_what_is_needed
     Conversion: convert_literal('false')
 Affected hosts: localhost

ERROR: Error while expanding attribute:
        Message: TemplatingError: An error occured while rendering a template (Component5): KeyError: 'doesnotexist'
      Attribute: Component5.attribute_cannot_be_expanded
 Affected hosts: localhost

ERROR: Need port for service address.
      Attribute: DNSProblem.attribute_with_problem
     Conversion: Address('localhost')
 Affected hosts: localhost

ERROR: Mode-string should be between `---------` and `rwxrwxrwx`.
      Attribute: FileMode > File('work/filemode/new-file.txt') > Mode('new-file.txt').mode
     Conversion: convert_mode('wrongmode')
 Affected hosts: localhost

ERROR: Overrides for undefined attributes
      Component: Component2
     Attributes: this_does_not_exist, this_also_does_not_exist
 Affected hosts: localhost

ERROR: Unused provided resources
    Resource "backend" provided by component3 with value ['192.168.0.1']
    Resource "frontend" provided by component3 with value ['test00.gocept.net']
    Resource "sub" provided by component3 with value [<SubComponent (localhost) "Component3 > SubComponent('sub sub')">]

ERROR: Unsatisfied resource requirements
    Resource "application" required by component4

ERROR: Found dependency cycle
cycle1 depends on
        cycle2
cycle2 depends on
        cycle1

ERROR: 10 remaining unconfigured component(s): component1, component2, component4, component5, crontab, cycle1, cycle2, dnsproblem, dnsproblem2, filemode
======================= 11 ERRORS - CONFIGURATION FAILED =======================
====================== DEPLOYMENT FAILED (during connect) ======================
"""
    )  # noqa: E501 line too long


def test_example_errors_missing_environment():
    os.chdir("examples/errors")
    out, _ = cmd("./batou deploy production", acceptable_returncodes=[1])
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `production`...

ERROR: Missing environment
    Environment: production
======================= DEPLOYMENT FAILED (during load) ========================
"""
    )  # noqa: E501 line too long


def test_example_ignores():
    os.chdir("examples/ignores")
    out, _ = cmd("./batou deploy ignores")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `ignores`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/2)
otherhost: Connection ignored (2/2)
================================== Deploying ===================================
localhost: Scheduling component component1 ...
localhost: Skipping component fail ... (Component ignored)
otherhost: Skipping component fail2 ... (Host ignored)
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=...s
============================= DEPLOYMENT FINISHED ==============================
"""
    )  # noqa: E501 line too long


def test_example_async_sync_deployment():
    os.chdir("examples/sync_async")
    out, _ = cmd("./batou -d deploy default")
    print(out)
    assert "Number of jobs: 1" in out

    out, _ = cmd("./batou -d deploy -j 2 default")
    print(out)
    assert "Number of jobs: 2" in out

    out, _ = cmd("./batou -d deploy async")
    print(out)
    assert "Number of jobs: 2" in out


def test_example_job_option_overrides_environment():
    os.chdir("examples/sync_async")
    out, _ = cmd("./batou -d deploy -j 5 async")
    print(out)
    assert "Number of jobs: 5" in out


def test_consistency_does_not_start_deployment():
    os.chdir("examples/tutorial-helloworld")
    out, _ = cmd("./batou deploy -c tutorial")
    print(out)
    assert "Deploying" not in out
    assert "localhost: Scheduling" not in out
    assert "CONSISTENCY CHECK FINISHED" in out

    # Counter-test that "Deploying" and "localhost: Scheduling" show
    # up if deployments happen
    out, _ = cmd("./batou deploy tutorial")
    print(out)
    assert "Deploying" in out
    assert "localhost: Scheduling" in out
    assert "CONSISTENCY CHECK FINISHED" not in out


def test_diff_is_not_shown_for_keys_in_secrets(tmp_path, monkeypatch, capsys):
    """It does not render diffs for files which contain secrets.

    Secrets might be in the config file in secrets/ or additional encrypted
    files belonging to the environment.
    """
    monkeypatch.chdir("examples/tutorial-secrets")
    if os.path.exists("work"):
        shutil.rmtree("work")
    try:
        out, _ = cmd("./batou deploy tutorial")
    finally:
        if os.path.exists("work"):
            shutil.rmtree("work")
        # shutil.rmtree("work")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `tutorial`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/1)
================================== Deploying ===================================
localhost: Scheduling component hello ...
localhost > Hello > File('work/hello/hello') > Presence('hello')
localhost > Hello > File('work/hello/hello') > Content('hello')
Not showing diff as it contains sensitive data,
see ...diff for the diff.
localhost > Hello > File('work/hello/other-secrets.yaml') > Presence('other-secrets.yaml')
localhost > Hello > File('work/hello/other-secrets.yaml') > Content('other-secrets.yaml')
Not showing diff as it contains sensitive data,
see ...diff for the diff.
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=...s
============================= DEPLOYMENT FINISHED ==============================
"""
    )  # noqa: E501 line too long


def test_diff_for_keys_in_secrets_overridable(tmp_path, monkeypatch, capsys):
    """It respects the "sensitive_data" flag when showing diffs for
    files which contain secrets

    Secrets might be in the config file in secrets/ or additional
    encrypted files belonging to the environment.
    """

    monkeypatch.chdir("examples/sensitive-values")
    if os.path.exists("work"):
        shutil.rmtree("work")
    try:
        out, _ = cmd("./batou deploy local")
    finally:
        if os.path.exists("work"):
            shutil.rmtree("work")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `local`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/1)
================================== Deploying ===================================
localhost: Scheduling component sensitivevalues ...
localhost > SensitiveValues > File('work/sensitivevalues/client_ed25519.key') > Presence('client_ed25519.key')
localhost > SensitiveValues > File('work/sensitivevalues/client_ed25519.key') > Content('client_ed25519.key')
Not showing diff as it contains sensitive data,
see ...diff for the diff.
localhost > SensitiveValues > File('work/sensitivevalues/client_ed25519.pub') > Presence('client_ed25519.pub')
localhost > SensitiveValues > File('work/sensitivevalues/client_ed25519.pub') > Content('client_ed25519.pub')
Not showing diff as it contains sensitive data,
see ...diff for the diff.
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_auto_rsa.pub') > Presence('hostkey_sensitive_auto_rsa.pub')
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_auto_rsa.pub') > Content('hostkey_sensitive_auto_rsa.pub')
  hostkey_sensitive_auto_rsa.pub ---
  hostkey_sensitive_auto_rsa.pub +++
  hostkey_sensitive_auto_rsa.pub @@ -0,0 +1 @@
  hostkey_sensitive_auto_rsa.pub +ssh-rsa ... batou-example-host
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_auto_ed25519.pub') > Presence('hostkey_sensitive_auto_ed25519.pub')
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_auto_ed25519.pub') > Content('hostkey_sensitive_auto_ed25519.pub')
Not showing diff as it contains sensitive data,
see ...diff for the diff.
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_masked_rsa.pub') > Presence('hostkey_sensitive_masked_rsa.pub')
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_masked_rsa.pub') > Content('hostkey_sensitive_masked_rsa.pub')
Not showing diff as it contains sensitive data.
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_clear_ed25519.pub') > Presence('hostkey_sensitive_clear_ed25519.pub')
localhost > SensitiveValues > File('work/sensitivevalues/hostkey_sensitive_clear_ed25519.pub') > Content('hostkey_sensitive_clear_ed25519.pub')
  hostkey_sensitive_clear_ed25519.pub ---
  hostkey_sensitive_clear_ed25519.pub +++
  hostkey_sensitive_clear_ed25519.pub @@ -0,0 +1 @@
  hostkey_sensitive_clear_ed25519.pub +ssh-ed25519 ... batou-example-host
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=...s
============================= DEPLOYMENT FINISHED ==============================
"""
    )


def test_durations_are_shown_for_components():
    os.chdir("examples/durations")
    out, _ = cmd("./batou deploy default")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `default`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/1)
================================== Deploying ===================================
localhost: Scheduling component takeslongtime ...
localhost > Takeslongtime [total=...s, verify=...s, update=NaN, sub=NaN]
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=...s
============================= DEPLOYMENT FINISHED ==============================
"""
    )


def test_check_consistency_works():
    os.chdir("examples/tutorial-secrets")
    out, _ = cmd("./batou deploy tutorial --consistency-only")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `tutorial`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/1)
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=NaN
========================== CONSISTENCY CHECK FINISHED ==========================
"""
    )


def test_predicting_deployment_works():
    os.chdir("examples/tutorial-secrets")
    out, _ = cmd("./batou deploy tutorial --predict-only")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `tutorial`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
localhost: Connecting via local (1/1)
======================== Predicting deployment actions =========================
localhost: Scheduling component hello ...
localhost > Hello > File('work/hello/hello') > Presence('hello')
localhost > Hello > File('work/hello/hello') > Content('hello')
Not showing diff as it contains sensitive data,
see .../examples/tutorial-secrets/work/.batou-diffs/...diff for the diff.
localhost > Hello > File('work/hello/other-secrets.yaml') > Presence('other-secrets.yaml')
localhost > Hello > File('work/hello/other-secrets.yaml') > Content('other-secrets.yaml')
Not showing diff as it contains sensitive data,
see .../examples/tutorial-secrets/work/.batou-diffs/...diff for the diff.
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=...s
======================== DEPLOYMENT PREDICTION FINISHED ========================
"""
    )


def test_check_consistency_works_with_local():
    os.chdir("examples/tutorial-secrets")
    out, _ = cmd("./batou deploy gocept --consistency-only --local")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `gocept`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
test01: Connecting via local (1/2)
test02: Connecting via local (2/2)
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=NaN
====================== CONSISTENCY CHECK (local) FINISHED ======================
"""
    )


def test_predicting_deployment_works_with_local():
    os.chdir("examples/tutorial-secrets")
    out, _ = cmd("./batou deploy gocept --predict-only --local")
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `gocept`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================
test01: Connecting via local (1/2)
test02: Connecting via local (2/2)
======================== Predicting deployment actions =========================
test01: Scheduling component hello ...
test02: Scheduling component hello ...
test01 > Hello > File('work/hello/hello') > Presence('hello')
test01 > Hello > File('work/hello/hello') > Content('hello')
  hello ---
  hello +++
  hello @@ -0,0 +1,2 @@
  hello +The magic word is None.
  hello +The other word is None.
test01 > Hello > File('work/hello/other-secrets.yaml') > Presence('other-secrets.yaml')
test01 > Hello > File('work/hello/other-secrets.yaml') > Content('other-secrets.yaml')
Not showing diff as it contains sensitive data,
see .../batou/examples/tutorial-secrets/work/.batou-diffs/...diff for the diff.
test02 > Hello > File('work/hello/hello') > Presence('hello')
test02 > Hello > File('work/hello/hello') > Content('hello')
  hello ---
  hello +++
  hello @@ -0,0 +1,2 @@
  hello +The magic word is None.
  hello +The other word is None.
test02 > Hello > File('work/hello/other-secrets.yaml') > Presence('other-secrets.yaml')
test02 > Hello > File('work/hello/other-secrets.yaml') > Content('other-secrets.yaml')
Not showing diff as it contains sensitive data,
see .../batou/examples/tutorial-secrets/work/.batou-diffs/...diff for the diff.
=================================== Summary ====================================
Deployment took total=...s, connect=...s, deploy=...s
==================== DEPLOYMENT PREDICTION (local) FINISHED ====================
"""
    )



import batou.lib.logrotate
import batou.lib.service
import batou.lib.supervisor
from batou.component import Attribute, Component, platform
from batou.lib.cron import CronJob


@platform("debian", batou.lib.service.Service)
class RebootCronjob(Component):
    def configure(self):
        self += CronJob(
            self.expand(
                "{{component.root.workdir}}/{{component.parent.executable}}"
            ),
            timing="@reboot",
            logger=self.root.name,
        )


# XXX can't use @platform since that's too late (see #12418)
class Supervisor(batou.lib.supervisor.Supervisor):

    pidfile = Attribute(default="var/supervisord.pid", map=True)


class Logrotate(batou.lib.logrotate.Logrotate):

    common_config = b"""\
daily
rotate 14
create
dateext
compress
notifempty
nomail
noolddir
missingok
sharedscripts
"""


@platform("debian", Logrotate)
class LogrotateCronjob(Component):
    def configure(self):
        self.directory = self.parent.workdir
        self += CronJob(
            self.expand(
                "/usr/sbin/logrotate -s {{component.directory}}/state"
                " {{component.directory}}/logrotate.conf"
            ),
            timing="45 2 * * *",
        )



import os.path
import urllib.parse

import requests

import batou
import batou.utils
from batou.component import Component

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib.request import urlretrieve


class Download(Component):

    namevar = "uri"

    target = None  # Filename where the download will be stored.
    checksum = None
    requests_kwargs = None

    def configure(self):
        if not self.target:
            self.target = urllib.parse.urlsplit(self.uri).path.split("/")[-1]
        if not self.target:
            raise KeyError(
                "No target is given and the URI does not allow "
                "deriving a filename."
            )
        if not self.checksum:
            raise ValueError("No checksum given.")
        self.checksum_function, self.checksum = self.checksum.split(":")

    def verify(self):
        if not os.path.exists(self.target):
            raise batou.UpdateNeeded()
        if self.checksum != batou.utils.hash(
            self.target, self.checksum_function
        ):
            raise batou.UpdateNeeded()

    def update(self):
        scheme = urllib.parse.urlsplit(self.uri)[0]
        if scheme in ["http", "https"]:
            self._update_requests()
        else:
            self._update_urllib()

        target_checksum = batou.utils.hash(self.target, self.checksum_function)
        assert (
            self.checksum == target_checksum
        ), """\
Checksum mismatch!
expected: %s
got: %s""" % (
            self.checksum,
            target_checksum,
        )

    def _update_requests(self):
        r = requests.get(
            self.uri, **(self.requests_kwargs if self.requests_kwargs else {})
        )
        r.raise_for_status()

        with open(self.target, "wb") as fd:
            for chunk in r.iter_content(4 * 1024**2):
                fd.write(chunk)

    def _update_urllib(self):
        path, headers = urlretrieve(self.uri, self.target)
        assert path == self.target

    @property
    def namevar_for_breadcrumb(self):
        uri = self.uri
        password = urllib.parse.urlsplit(uri).password
        if password:
            # Indeed, this is a very simple approach.
            uri = uri.replace(password, "*****")
        return uri



import ast
import os
import subprocess
import sys

import execnet.gateway_io
import yaml

from batou import DeploymentError, SilentConfigurationError, output, remote_core
from batou.utils import BagOfAttributes

# Keys in os.environ which get propagated to the remote side:
REMOTE_OS_ENV_KEYS = (
    "REMOTE_PDB_HOST",
    "REMOTE_PDB_PORT",
)


# Monkeypatch execnet to support 'vagrant ssh' and 'kitchen exec'.
# 'vagrant' support has been added to 'execnet' release 1.4.
def get_kitchen_ssh_connection_info(name):
    cmd = "kitchen", "diagnose", "--log-level=error", name
    info = yaml.load(subprocess.check_output(cmd))
    (instance,) = list(info["instances"].values())
    state = instance["state_file"]
    return [
        "-o",
        "StrictHostKeyChecking=no",
        "-i",
        state["ssh_key"],
        "-p",
        state["port"],
        "-l",
        state["username"],
        state["hostname"],
    ]


def new_ssh_args(spec):
    from execnet.gateway_io import popen_bootstrapline

    remotepython = spec.python or "python"
    if spec.type == "vagrant":
        args = ["vagrant", "ssh", spec.ssh, "--", "-C"]
    elif spec.type == "kitchen":
        # TODO: this should really use:
        #   args = ['kitchen', 'exec', spec.ssh, '-c']
        # but `exec` apparently doesn't connect stdin (yet)...
        args = ["ssh", "-C"] + get_kitchen_ssh_connection_info(spec.ssh)
    else:
        args = ["ssh", "-C"]
    if spec.ssh_config is not None:
        args.extend(["-F", str(spec.ssh_config)])
    remotecmd = '%s -c "%s"' % (remotepython, popen_bootstrapline)
    if spec.type == "vagrant" or spec.type == "kitchen":
        args.extend([remotecmd])
    else:
        args.extend([spec.ssh, remotecmd])
    return args


execnet.gateway_io.ssh_args = new_ssh_args


class RPCWrapper(object):
    def __init__(self, host):
        self.host = host

    def __getattr__(self, name):
        def call(*args, **kw):
            output.annotate(
                "rpc {}: {}(*{}, **{})".format(self.host.fqdn, name, args, kw),
                debug=True,
            )
            self.host.channel.send((name, args, kw))
            while True:
                message = self.host.channel.receive()
                output.annotate(
                    "{}: message: {}".format(self.host.fqdn, message),
                    debug=True,
                )
                type = message[0]
                if type == "batou-result":
                    return message[1]
                elif type == "batou-output":
                    _, output_cmd, args, kw = message
                    getattr(output, output_cmd)(*args, **kw)
                elif type == "batou-unknown-error":
                    output.error(message[1])
                    raise RuntimeError(
                        "{}: Remote exception encountered.".format(
                            self.host.fqdn
                        )
                    )
                elif type == "batou-error":
                    # Remote put out the details already.
                    raise RuntimeError(
                        "{}: Remote exception encountered.".format(
                            self.host.fqdn
                        )
                    )
                else:
                    raise RuntimeError(
                        "{}: Unknown message type {}".format(
                            self.host.fqdn, type
                        )
                    )

        return call


_no_value_marker = object()


class Host(object):

    service_user = None
    require_sudo = None
    ignore = False
    platform = None
    _provisioner = None
    remap = False
    ignore = False

    def __init__(self, name, environment, config={}):
        # The _name attribute is the name that is given to this host in the
        # environment. The `name` property will return the true name for this
        # host in case that a mapping exists, e.g. due to a provisioner.
        self._name = name

        self.aliases = BagOfAttributes()

        self.data = {}

        self.rpc = RPCWrapper(self)
        self.environment = environment

        self.ignore = ast.literal_eval(config.get("ignore", "False"))

        self.platform = config.get("platform", environment.platform)
        self.service_user = config.get("service_user", environment.service_user)
        if "require_sudo" in config:
            self.require_sudo = ast.literal_eval(config.get("require_sudo"))
        else:
            self.require_sudo = environment.require_sudo

        self.remap = ast.literal_eval(
            config.get("provision-dynamic-hostname", "False")
        )
        self._provisioner = config.get("provisioner")
        if self.provisioner:
            self.provisioner.configure_host(self, config)

        for key, value in list(config.items()):
            if key.startswith("data-"):
                key = key.replace("data-", "", 1)
                self.data[key] = value

    @property
    def provisioner(self):
        if self._provisioner == "none":
            # Provisioning explicitly disabled for this host
            return
        elif not self._provisioner:
            # Default provisionier (if available)
            return self.environment.provisioners.get("default")
        return self.environment.provisioners[self._provisioner]

    # These are internal aliases to allow having an explicit name
    # for a host in the environment and then having a provisioner assign a
    # different "true" name for this host.
    @property
    def _aliases(self):
        if self._name == self.name:
            return []
        return [self._name]

    @property
    def name(self):
        mapping = self.environment.hostname_mapping
        if not self.remap:
            # Update the map in case it contained an old persisted entry.
            mapping[self._name] = self._name
        elif self._name not in mapping:
            mapping[self._name] = self.provisioner.suggest_name(self._name)
        return mapping[self._name]

    @property
    def fqdn(self):
        name = self.name
        if self.environment.host_domain:
            name += "." + self.environment.host_domain
        return name

    def deploy_component(self, component, predict_only):
        self.rpc.deploy(component, predict_only)

    def root_dependencies(self):
        return self.rpc.root_dependencies()

    @property
    def components(self):
        return self.environment.components_for(self)

    def summarize(self):
        if self.provisioner:
            self.provisioner.summarize(self)


class LocalHost(Host):
    def connect(self):
        self.gateway = execnet.makegateway(
            "popen//python={}".format(sys.executable)
        )
        self.channel = self.gateway.remote_exec(remote_core)

    def start(self):
        self.rpc.lock()

        # Since we reconnected, any state on the remote side has been lost,
        # so we need to set the target directory again (which we only can
        # know about locally).
        self.rpc.setup_output(output.enable_debug)

        env = self.environment

        self.remote_repository = self.rpc.ensure_repository(
            env.target_directory, "local"
        )

        self.remote_base = self.rpc.ensure_base(env.deployment_base)

        # XXX the cwd isn't right.
        return self.rpc.setup_deployment(
            env.name,
            self.name,
            env.overrides,
            env.secret_files,
            env.secret_data,
            env._host_data(),
            env.timeout,
            env.platform,
        )

    def disconnect(self):
        if hasattr(self, "gateway"):
            self.gateway.exit()


class RemoteHost(Host):

    gateway = None

    def _makegateway(self, interpreter):
        if self.service_user is not None and self.require_sudo:
            # When calling sudo, ensure that no password will ever be
            # requested, and fail otherwise.
            interpreter = "sudo -ni -u {user} {interpreter}".format(
                user=self.service_user, interpreter=interpreter
            )
        spec = "ssh={fqdn}//python={interpreter}//type={method}".format(
            fqdn=self.fqdn,
            method=self.environment.connect_method,
            interpreter=interpreter,
        )
        ssh_configs = [
            "ssh_config_{}".format(self.environment.name),
            "ssh_config",
        ]
        for ssh_config in ssh_configs:
            if os.path.exists(ssh_config):
                spec += "//ssh_config={}".format(ssh_config)
                break

        return execnet.makegateway(spec)

    def connect(self, interpreter="python3"):
        if self.gateway:
            output.annotate("Disconnecting ...", debug=True)
            self.disconnect()

        output.annotate("Connecting ...", debug=True)

        self.gateway = self._makegateway(interpreter)

        try:
            self.channel = self.gateway.remote_exec(remote_core)
        except IOError:
            raise RuntimeError(
                "Could not start batou on host `{}`. "
                "The output above may contain more information. ".format(
                    self.fqdn
                )
            )

        if self.service_user is not None and self.require_sudo is None:
            # Discover whether we need to invoke sudo to reach the
            # right user.
            remote_user = self.rpc.whoami()
            self.require_sudo = remote_user != self.service_user

            if self.require_sudo:
                output.annotate(
                    "Service user requires sudo, reconnecting ...", debug=True
                )

                self.gateway.exit()
                self.gateway = self._makegateway(interpreter)

                try:
                    self.channel = self.gateway.remote_exec(remote_core)
                except IOError:
                    raise RuntimeError(
                        "Could not start batou on host `{}`. "
                        "The output above may contain more information. ".format(
                            self.fqdn
                        )
                    )

        output.annotate("Connected ...", debug=True)

    def start(self):
        output.step(self.name, "Bootstrapping ...", debug=True)
        self.rpc.lock()
        env = self.environment

        self.remote_repository = self.rpc.ensure_repository(
            env.target_directory, env.update_method
        )
        self.remote_base = self.rpc.ensure_base(env.deployment_base)

        output.step(self.name, "Updating repository ...", debug=True)
        env.repository.update(self)

        self.rpc.build_batou()

        # Now, replace the basic interpreter connection, with a "real" one
        # that has all our dependencies installed.
        #
        # XXX this requires an interesting move of detecting which appenv
        # version we have available to make this backwards compatible.
        self.connect(self.remote_base + "/appenv python")

        # Reinit after reconnect ...
        self.rpc.lock()
        self.remote_repository = self.rpc.ensure_repository(
            env.target_directory, env.update_method
        )
        self.remote_base = self.rpc.ensure_base(env.deployment_base)

        # Since we reconnected, any state on the remote side has been lost,
        # so we need to set the target directory again (which we only can
        # know about locally)
        self.rpc.setup_output(output.enable_debug)

        return self.rpc.setup_deployment(
            env.name,
            self.name,
            env.overrides,
            env.secret_files,
            env.secret_data,
            env._host_data(),
            env.timeout,
            env.platform,
            {
                key: os.environ.get(key)
                for key in REMOTE_OS_ENV_KEYS
                if os.environ.get(key)
            },
        )

    def disconnect(self):
        if self.gateway is not None:
            self.gateway.exit()
